"""Top-level package for bib-dedupe."""

__author__ = """Gerit Wagner"""
__email__ = "gerit.wagner@uni-bamberg.de"

# Instead of adding elements to __all__,
# prefixing methods/variables with "__" is preferred.
# Imports like "from x import *" are discouraged.

from bib_dedupe.util import VerbosePrint

verbosity_level = 1  # Set your desired verbosity level here
verbose_print = VerbosePrint(verbosity_level)



#! /usr/bin/env python
"""Preparation of number field"""
import re

import numpy as np


def prep_number(number_array: np.array) -> np.array:
    number_array = np.array(
        [re.sub(r"[A-Za-z.]*", "", number) for number in number_array]
    )
    number_array = np.array(
        [
            # pages included: "6 51-6"
            re.search(r"(\d+) .*", number).group(1)  # type: ignore
            if re.search(r"(\d+) \d+-\d+", number) is not None
            else number
            for number in number_array
        ]
    )

    number_array = np.array(
        [
            number.replace(" ", "").replace("(", "").replace(")", "")
            for number in number_array
        ]
    )

    return np.array(
        ["" if number in ["nan", "var.pagings"] else number for number in number_array]
    )



#! /usr/bin/env python
"""Module for preparing bibliographic data for deduplication."""
import concurrent.futures
import os
import time
from datetime import datetime

import numpy as np
import pandas as pd

from bib_dedupe import verbose_print
from bib_dedupe.constants.fields import ABSTRACT
from bib_dedupe.constants.fields import AUTHOR
from bib_dedupe.constants.fields import BOOKTITLE
from bib_dedupe.constants.fields import CONTAINER_TITLE
from bib_dedupe.constants.fields import DOI
from bib_dedupe.constants.fields import ENTRYTYPE
from bib_dedupe.constants.fields import ID
from bib_dedupe.constants.fields import JOURNAL
from bib_dedupe.constants.fields import NUMBER
from bib_dedupe.constants.fields import PAGES
from bib_dedupe.constants.fields import SEARCH_SET
from bib_dedupe.constants.fields import SERIES
from bib_dedupe.constants.fields import TITLE
from bib_dedupe.constants.fields import VOLUME
from bib_dedupe.constants.fields import YEAR
from bib_dedupe.prep_abstract import prep_abstract
from bib_dedupe.prep_author import prep_authors
from bib_dedupe.prep_author import select_authors
from bib_dedupe.prep_container_title import get_container_title_short
from bib_dedupe.prep_container_title import prep_container_title
from bib_dedupe.prep_container_title import set_container_title
from bib_dedupe.prep_doi import prep_doi
from bib_dedupe.prep_number import prep_number
from bib_dedupe.prep_pages import prep_pages
from bib_dedupe.prep_title import prep_title
from bib_dedupe.prep_volume import prep_volume
from bib_dedupe.prep_year import prep_year

pd.set_option('future.no_silent_downcasting', True)

REQUIRED_FIELDS = [ID, ENTRYTYPE, TITLE, AUTHOR, YEAR]
OPTIONAL_FIELDS = [
    JOURNAL,
    BOOKTITLE,
    SERIES,
    VOLUME,
    NUMBER,
    PAGES,
    ABSTRACT,
    DOI,
    SEARCH_SET,
]
ALL_FIELDS = REQUIRED_FIELDS + OPTIONAL_FIELDS + [CONTAINER_TITLE]

function_mapping = {
    AUTHOR: prep_authors,
    TITLE: prep_title,
    CONTAINER_TITLE: prep_container_title,
    YEAR: prep_year,
    VOLUME: prep_volume,
    NUMBER: prep_number,
    PAGES: prep_pages,
    ABSTRACT: prep_abstract,
    DOI: prep_doi,
}


def prepare_df_split(split_df: pd.DataFrame) -> pd.DataFrame:
    """
    Prepare a split dataframe for deduplication.

    Args:
        split_df: A split dataframe.

    Returns:
        The processed dataframe.
    """
    split_df.replace(
        to_replace={
            "UNKNOWN": "",
            "n/a": "",
            "N/A": "",
            "NA": "",
            "&amp;": "and",
            " & ": " and ",
            " + ": " and ",
        },
        inplace=True,
    )

    set_container_title(split_df)

    split_df["author_full"] = split_df[AUTHOR]

    for field, function in function_mapping.items():
        split_df[field] = function(split_df[field].values)  # type: ignore

    split_df[AUTHOR] = select_authors(split_df[AUTHOR].values)

    # Fix cases where years are erroneously entered into pages field
    split_df.loc[split_df[PAGES] == split_df[YEAR], PAGES] = ""

    if BOOKTITLE in split_df.columns:
        split_df = split_df.drop(columns=[BOOKTITLE])
    if JOURNAL in split_df.columns:
        split_df = split_df.drop(columns=[JOURNAL])

    split_df = split_df.fillna("")

    return split_df


def __general_prep(records_df: pd.DataFrame) -> pd.DataFrame:
    """
    Perform general preparation on the records dataframe.

    Args:
        records_df: The records dataframe.

    Returns:
        The prepared dataframe.
    """
    records_df = records_df.copy()

    if ID not in records_df.columns:
        records_df.loc[:, ID] = range(1, len(records_df) + 1)
    if not records_df[ID].is_unique:
        raise ValueError("ID column in records_df must be unique.")

    if ENTRYTYPE not in records_df.columns:
        records_df[ENTRYTYPE] = "article"

    missing_fields = [f for f in REQUIRED_FIELDS if f not in records_df.columns]
    assert len(missing_fields) == 0, f"Missing required fields: {missing_fields}"

    for column in records_df.columns:
        records_df[column] = (
            records_df[column]
            .replace(["#NAME?", "UNKNOWN", ""], np.nan)
        )
    if records_df[TITLE].isnull().any():
        verbose_print.print(
            "Warning: Some records have empty title field. These records will not be considered."
        )
        records_df = records_df.dropna(subset=[TITLE])

    # if columns are of type float, we need to avoid casting "3.0" to "30"
    for col in records_df.columns:
        if records_df[col].dtype == float:
            records_df[col] = records_df[col].apply(
                lambda x: str(int(x))
                if pd.notna(x) and isinstance(x, (int, float))
                else ""
            )

    for optional_field in OPTIONAL_FIELDS:
        if optional_field not in records_df.columns:
            records_df = records_df.assign(**{optional_field: ""})

    records_df = records_df.drop(
        labels=list(records_df.columns.difference(ALL_FIELDS)),
        axis=1,
    )
    records_df.loc[:, CONTAINER_TITLE] = ""
    records_df.loc[:, ALL_FIELDS] = records_df[ALL_FIELDS].astype(str)

    return records_df


def determine_cpu_count(requested_cpu: int, record_count: int) -> int:
    """
    Determines the number of CPUs to use based on the requested CPU count and the number of records.

    Args:
        requested_cpu: The requested CPU count.
        record_count: The number of records.

    Returns:
        The number of CPUs to use.
    """
    if requested_cpu == -1:
        cpu_count = os.cpu_count() or 1
    else:
        cpu_count = requested_cpu

    # For small datasets, use a single CPU
    if record_count < 100:
        cpu_count = 1

    return cpu_count


def prep(records_df: pd.DataFrame, *, cpu: int = -1) -> pd.DataFrame:
    """
    Prepare records for deduplication.

    Args:
        records_df: The records dataframe.
        cpu: The number of CPUs to use. If -1, use all available CPUs.

    Returns:
        The prepared records dataframe.
    """

    now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    verbose_print.print(f"Loaded {records_df.shape[0]:,} records")
    verbose_print.print(f"Prep started at {now}")
    start_time = time.time()

    if 0 == records_df.shape[0]:
        verbose_print.print("No records to prepare")
        return {}

    records_df = __general_prep(records_df)
    cpu = determine_cpu_count(cpu, records_df.shape[0])
    if cpu == 1:
        records_df = prepare_df_split(records_df)
    else:
        df_split = np.array_split(records_df, cpu)
        with concurrent.futures.ProcessPoolExecutor(max_workers=cpu) as executor:
            results = executor.map(prepare_df_split, df_split)
        records_df = pd.concat(list(results))
    records_df = records_df.assign(
        author_first=records_df[AUTHOR].str.split().str[0],
        title_short=records_df[TITLE].apply(lambda x: " ".join(x.split()[:10])),
        container_title_short=get_container_title_short(
            records_df[CONTAINER_TITLE].values
        ),
    )

    for column in records_df.columns:
        records_df.loc[records_df[column] == "nan", column] = ""

    end_time = time.time()
    verbose_print.print(f"Prep completed after: {end_time - start_time:.2f} seconds")

    return records_df



#! /usr/bin/env python
"""Preparation of title field"""
import html
import re

import numpy as np
from number_parser import parse


TITLE_STOPWORDS = [
    "a",
    "an",
    "the",
    "in",
    "of",
    "on",
    "for",
    "from",
    "does",
    "do",
    "and",
    "are",
    "on",
    "with",
    "to",
    "or",
    "as",
    "by",
    "their",
    "the",
]


def remove_erratum_suffix(title: str) -> str:
    erratum_phrases = ["erratum appears in ", "erratum in "]
    for phrase in erratum_phrases:
        if phrase in title.lower():
            title = title[: title.lower().rfind(phrase) - 2]

    title = re.sub(r" review \d+ refs$", "", title)

    return title


# flake8: noqa: E501
# pylint: disable=line-too-long
def prep_title(title_array: np.array) -> np.array:
    # Remove language and contents after if "Russian" or "Chinese" followed by a newline
    title_array = np.array(
        [
            re.sub(
                r"\. (Russian|Chinese|Spanish|Czech|Italian|Polish|Dutch|Ukrainian|German|French|Japanese|Slovak|Hungarian|Portuguese English|Turkish|Norwegian|Portuguese)(\r?\n)?.*$",
                "",
                title,
                flags=re.IGNORECASE,
            )
            if ". " in title
            else title
            for title in title_array
        ]
    )

    title_array = np.array(
        [
            title.replace("-like", "like")
            .replace("co-", "co")
            .replace("post-", "post")
            .replace("three-dimensional", "threedimensional")
            .replace("+", " plus ")
            for title in title_array
        ]
    )

    # Replace 'withdrawn' at the beginning and '(review') at the end (case insensitive)
    title_array = np.array(
        [
            re.sub(
                r"^(withdrawn[.:] )|^(proceedings: )|^(reprint)|( \(review\))$|( \(vol \d+.*\))",
                "",
                title,
                flags=re.IGNORECASE,
            )
            for title in title_array
        ]
    )

    # Replace roman numbers (title similarity is sensitive to numbers)
    title_array = np.array(
        [
            re.sub(
                r"\biv\b",
                " 4 ",
                re.sub(
                    r"\biii\b",
                    " 3 ",
                    re.sub(
                        r"\bii\b",
                        " 2 ",
                        re.sub(r"\bi\b", " 1 ", title, flags=re.IGNORECASE),
                        flags=re.IGNORECASE,
                    ),
                    flags=re.IGNORECASE,
                ),
                flags=re.IGNORECASE,
            )
            for title in title_array
        ]
    )

    # Remove html tags
    title_array = np.array([re.sub(r"<.*?>", " ", title) for title in title_array])
    # Replace html special entities
    title_array = np.array([html.unescape(title) for title in title_array])

    # Remove language tags added by some databases (at the end)
    title_array = np.array(
        [re.sub(r"\. \[[A-Z][a-z]*\]$", "", title) for title in title_array]
    )

    # Remove trailing "1" if it is not preceded by "part"
    title_array = np.array(
        [
            re.sub(r"1$", "", title) if "part" not in title[-10:].lower() else title
            for title in title_array
        ]
    )
    # Remove erratum suffix
    # https://www.nlm.nih.gov/bsd/policy/errata.html
    title_array = np.array([remove_erratum_suffix(title) for title in title_array])

    # Replace words in parentheses at the end
    title_array = np.array(
        [re.sub(r"\s*\([^)]*\)\s*$", "", value) for value in title_array]
    )

    # Remove '[Review] [33 refs]' and ' [abstract no: 134]' ignoring case
    title_array = np.array(
        [
            re.sub(
                r"\[Review\] \[\d+ refs\]| \[abstract no: \d+\]",
                "",
                title,
                flags=re.IGNORECASE,
            )
            for title in title_array
        ]
    )

    # Replace brackets (often used in formulae)
    title_array = np.array(
        [re.sub(r"([A-Za-z])\(([0-9]*)\)", r"\1\2", title) for title in title_array]
    )

    # Replace special characters
    title_array = np.array(
        [re.sub(r"[^A-Za-z0-9,\[\]]+", " ", title.lower()) for title in title_array]
    )

    # Remove common stopwords
    title_array = np.array(
        [
            " ".join(word for word in title.split() if word not in TITLE_STOPWORDS)
            for title in title_array
        ]
    )

    # Apply parse function to replace numbers
    title_array = np.array([parse(title) for title in title_array])

    # Replace spaces between digits
    title_array = np.array(
        [
            re.sub(r"(\d) (\d)", r"\1\2", title).rstrip(" ].").lstrip("[ ")
            for title in title_array
        ]
    )

    # Replace multiple spaces with a single space
    title_array = np.array(
        [re.sub(r"\s+", " ", title).rstrip().lstrip() for title in title_array]
    )
    return title_array



#! /usr/bin/env python
"""Preparation of doi field"""
import re
import urllib.parse

import numpy as np


def prep_doi(doi_array: np.array) -> np.array:
    doi_array = np.array(
        [re.sub(r"http://dx.doi.org/", "", doi.lower()) for doi in doi_array]
    )
    doi_array = np.array([re.sub(r"-", "_", doi) for doi in doi_array])
    doi_array = np.array([re.sub(r"\[doi\]", "", doi) for doi in doi_array])
    doi_array = np.array([re.sub(r"[\r\n]+", " ; ", doi) for doi in doi_array])

    doi_array = np.array(
        [
            doi.split(";")[1].lstrip()
            if ";" in doi and doi.split(";")[1].lstrip().startswith("10.")
            else doi.split(";")[0].lstrip()
            if ";" in doi and doi.split(";")[0].lstrip().startswith("10.")
            else doi
            for doi in doi_array
        ]
    )
    doi_array = np.array(
        [doi.split("[pii];")[1] if "[pii];" in doi else doi for doi in doi_array]
    )
    doi_array = np.array([urllib.parse.unquote(doi) for doi in doi_array])
    doi_array = np.array([doi if doi.startswith("10.") else "" for doi in doi_array])

    return np.array(
        ["" if doi == "nan" else doi.replace(".", "").rstrip() for doi in doi_array]
    )



#! /usr/bin/env python
"""Module for deduplicating bibliographic records"""
from __future__ import annotations

import os
import typing

import pandas as pd
import requests

import bib_dedupe.block
import bib_dedupe.match
import bib_dedupe.maybe_cases
import bib_dedupe.merge
import bib_dedupe.prep
import bib_dedupe.sim
from bib_dedupe import verbose_print


def prep(
    records_df: pd.DataFrame,
    *,
    verbosity_level: typing.Optional[int] = None,
    cpu: int = -1,
) -> pd.DataFrame:
    """Preprocesses records for deduplication.

    Args:
        records_df (pd.DataFrame): The dataframe containing the records to be preprocessed.
        verbosity_level (int, optional): Level of verbosity for logging. Defaults to None.

    Returns:
        pd.DataFrame: The preprocessed records dataframe.
    """
    if verbosity_level is not None:
        verbose_print.verbosity_level = verbosity_level
    return bib_dedupe.prep.prep(records_df, cpu=cpu)


def block(
    records_df: pd.DataFrame,
    *,
    verbosity_level: typing.Optional[int] = None,
    cpu: int = -1,
) -> pd.DataFrame:
    """
    Blocks pairs of records for deduplication.

    Args:
        records_df (pd.DataFrame): The dataframe containing the records to be deduplicated.
        verbosity_level (int, optional): Level of verbosity for logging. Defaults to None.

    Returns:
        pd.DataFrame: The dataframe containing the blocked pairs for deduplication.
    """
    if verbosity_level is not None:
        verbose_print.verbosity_level = verbosity_level

    pairs_df = bib_dedupe.block.block(records_df, cpu=cpu)

    return pairs_df


def match(
    pairs_df: pd.DataFrame,
    *,
    verbosity_level: typing.Optional[int] = None,
    cpu: int = -1,
) -> pd.DataFrame:
    """
    Identifies true/maybe matches from the given pairs based on similarity scores.

    Args:
        pairs_df (pd.DataFrame): The DataFrame containing the pairs to be compared.
        verbosity_level (int, optional): Level of verbosity for logging. Defaults to None.
        debug (bool, optional): If True, enables debug mode. Defaults to False.

    Returns:
        pd.DataFrame: The DataFrame containing the true/maybe matches.
    """
    if verbosity_level is not None:
        verbose_print.verbosity_level = verbosity_level

    return bib_dedupe.match.match(pairs_df, cpu=cpu)


def export_maybe(
    records_df: pd.DataFrame,
    *,
    matched_df: pd.DataFrame,
    verbosity_level: typing.Optional[int] = None,
) -> None:
    """
    Exports 'maybe' cases for manual review during deduplication.

    Args:
        matched_df (pd.DataFrame): The dataframe containing the matched pairs.
        records_df (pd.DataFrame): The dataframe containing the records.
        verbosity_level (int, optional): Level of verbosity for logging. Defaults to None.
    """
    if verbosity_level is not None:
        verbose_print.verbosity_level = verbosity_level

    bib_dedupe.maybe_cases.export_maybe(matched_df, records_df)


def import_maybe(
    matched_df: pd.DataFrame, *, verbosity_level: typing.Optional[int] = None
) -> pd.DataFrame:
    """
    Imports decisions for 'maybe' cases after manual review.

    Args:
        matched_df (pd.DataFrame): The dataframe containing the matches.
        verbosity_level (int, optional): Level of verbosity for logging. Defaults to None.

    Returns:
        pd.DataFrame: The dataframe containing the updated matches.
    """
    if verbosity_level is not None:
        verbose_print.verbosity_level = verbosity_level

    return bib_dedupe.maybe_cases.import_maybe(matched_df)


def cluster(
    matched_df: pd.DataFrame, *, verbosity_level: typing.Optional[int] = None
) -> list:
    """
    Clusters the matched data for deduplication.

    Args:
        matched_df (pd.DataFrame): The dataframe containing the matches.
        verbosity_level (int, optional): Level of verbosity for logging. Defaults to None.

    Returns:
        list: The list of clusters.
    """
    if verbosity_level is not None:
        verbose_print.verbosity_level = verbosity_level

    return bib_dedupe.cluster.get_connected_components(matched_df)


def merge(
    records_df: pd.DataFrame,
    *,
    matched_df: typing.Optional[pd.DataFrame] = None,
    duplicate_id_sets: typing.Optional[list] = None,
    verbosity_level: typing.Optional[int] = None,
) -> pd.DataFrame:
    """
    Merges duplicate records in the given dataframe.

    Args:
        records_df (pd.DataFrame): The DataFrame containing the records to be merged.
        duplicate_id_sets (list, optional): List of sets containing duplicate
        record IDs. If None, the function will perform deduplication process
        to identify duplicates. Defaults to None.
        verbosity_level (int, optional): Level of verbosity for logging. Defaults to None.

    Returns:
        pd.DataFrame: The merged DataFrame.
    """
    if verbosity_level is not None:
        verbose_print.verbosity_level = verbosity_level

    if matched_df:
        duplicate_id_sets = bib_dedupe.cluster.get_connected_components(matched_df)

    if not duplicate_id_sets:
        prep_df = prep(records_df)
        blocked_df = block(records_df=prep_df)
        matched_df = match(blocked_df)
        duplicate_id_sets = bib_dedupe.cluster.get_connected_components(matched_df)

    return bib_dedupe.merge.merge(records_df, duplicate_id_sets=duplicate_id_sets)


def _download_file_from_github(url: str, local_path: str) -> None:
    raw_url = url.replace("github.com", "raw.githubusercontent.com").replace(
        "/blob", ""
    )
    response = requests.get(raw_url, stream=True)
    with open(local_path, "wb") as file:
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                file.write(chunk)


def load_example_data(dataset: str) -> pd.DataFrame:
    """
    Retrieves the dataset from GitHub and loads the records as a DataFrame.

    Args:
        dataset (str): The name of the dataset (e.g., stroke, cardiac, depression).

    Returns:
        pd.DataFrame: The loaded records dataframe.
    """

    local_path = f"{dataset}.csv"
    url = f"https://github.com/CoLRev-Environment/bib-dedupe/blob/main/data/{dataset}/records_pre_merged.csv"

    if not os.path.exists(local_path):
        _download_file_from_github(url, local_path)

    try:
        df = pd.read_csv(local_path)
    except FileNotFoundError:
        raise ValueError(f"Dataset '{dataset}' not found.")
    return df



#! /usr/bin/env python
"""Matching rules for dedupe"""
from bib_dedupe.constants.fields import ABSTRACT
from bib_dedupe.constants.fields import AUTHOR
from bib_dedupe.constants.fields import CONTAINER_TITLE
from bib_dedupe.constants.fields import DOI
from bib_dedupe.constants.fields import ENTRYTYPE
from bib_dedupe.constants.fields import NUMBER
from bib_dedupe.constants.fields import PAGE_RANGES_ADJACENT
from bib_dedupe.constants.fields import PAGES
from bib_dedupe.constants.fields import TITLE
from bib_dedupe.constants.fields import VOLUME
from bib_dedupe.constants.fields import YEAR

# flake8: noqa: E501
# pylint: disable=line-too-long


def mismatch(*keys: str) -> str:
    return "&".join(
        f" ({key}_1 != {key}_2 & {key}_1 != '' & {key}_2 != '') " for key in keys
    )


def match(*args: str, threshold: float = 1.0) -> str:
    if threshold == 1.0:
        return "&".join(f" ({arg} == {threshold}) " for arg in args)
    return "&".join(f" ({arg} > {threshold}) " for arg in args)


def non_contradicting(*keys: str) -> str:
    return " & ".join(
        f" ( {key}_1 == {key}_2  | {key}_1 == '' | {key}_2 == '' ) " for key in keys
    )


def both_entrytypes(entrytype: str) -> str:
    return f"({ENTRYTYPE}_1 == '{entrytype}' & {ENTRYTYPE}_2 == '{entrytype}')"


au07_ti10_ct10 = f" {match(TITLE, CONTAINER_TITLE)} & {AUTHOR} > 0.7 "
au10_ti07_ct10 = f" ({match(AUTHOR, CONTAINER_TITLE)} & {TITLE} > 0.7 ) "
au10_ti10_ct07 = f" {match(TITLE, AUTHOR)} & {CONTAINER_TITLE} > 0.7 "
au095_ti09_ct075 = f" ({AUTHOR} > 0.95 & {TITLE} > 0.9 &  {CONTAINER_TITLE} > 0.75) "
au08_ti09_ct09 = f" ({AUTHOR} > 0.8 & {TITLE} > 0.9 & {CONTAINER_TITLE} > 0.9) "
au09_ti09_ctXX = f" ({AUTHOR} > 0.9 & {TITLE} > 0.9) "
auXX_ti095_ct095 = f" ({TITLE} > 0.95 & {CONTAINER_TITLE} > 0.95) "
au10_tiXX_ct10 = f" ({match(AUTHOR)} & {match(CONTAINER_TITLE)}) "
au10_ti10_ctNC = f" {match(AUTHOR, TITLE)} & {non_contradicting(CONTAINER_TITLE)} "

# Notes:
# - structure of conditions:
#   - author/title/container-title first (they require continuous similarities, they rarely have missing data, and field-misassignments are very rare (e.g., author in title field))
#   - volume/number/pages/year/doi/abstract (most require binary similarities, they can have missing data (see non_contradicting()), and misplacements are possible (e.g., number in volume field))
#   - VOLUME, NUMBER, PAGES, generally be non_contradicting (because they have missing values). similarities of 0.8/... may not be very helpful considering that we mainly have similarities of 0 or 1
#   - especially for doi and abstract, non_contradicting is preferred.

# - "contained_in ..." is not supported by query(). -> covered in similarities
# - Queries are better for debugging (understanding which conditions do/do not apply)
#   https://jakevdp.github.io/PythonDataScienceHandbook/03.12-performance-eval-and-query.html

duplicate_conditions = [
    # Substantial differences in one of AUTHOR/TITLE/CONTAINER_TITLE
    f"({au07_ti10_ct10} & {match(VOLUME, PAGES)})",
    f"({au07_ti10_ct10} & {non_contradicting(VOLUME, NUMBER, PAGES, YEAR, DOI)})",
    f"({au10_ti07_ct10} & {non_contradicting(NUMBER, PAGES, YEAR, DOI)})",
    f"({au10_ti10_ct07} & {non_contradicting(VOLUME, NUMBER, PAGES, YEAR, DOI)})",
    # Differences across AUTHOR/TITLE/CONTAINER_TITLE
    f"({au08_ti09_ct09} & {non_contradicting(VOLUME, NUMBER, YEAR, DOI)} & {PAGES} > 0.75 )",
    f"({au08_ti09_ct09} & {non_contradicting(VOLUME, NUMBER, PAGES, DOI)})",
    f"({au095_ti09_ct075} & {non_contradicting(VOLUME, NUMBER, PAGES, YEAR, DOI)})",
    f"({au095_ti09_ct075} & {match(NUMBER, PAGES)})",
    f"({au095_ti09_ct075} & {match(VOLUME, NUMBER)})",
    f"({au095_ti09_ct075} & {match(VOLUME, PAGES)})",
    f"({au095_ti09_ct075} & {match(VOLUME)} & {ABSTRACT} > 0.9)",
    f"({au095_ti09_ct075} & {match(YEAR, ABSTRACT)})",
    # Special cases
    f'({au095_ti09_ct075} & {both_entrytypes("inproceedings")} & {match(YEAR)})',  # Inproceedings
    f"({au07_ti10_ct10} & {DOI} > 0.9)",  # Updates
    # no AUTHOR
    f"({auXX_ti095_ct095} & {non_contradicting(VOLUME, NUMBER, PAGES, YEAR, DOI)})",
    f"({auXX_ti095_ct095} & {match(VOLUME, NUMBER, PAGES, YEAR)} & {non_contradicting(DOI, ABSTRACT)})",
    # no CONTAINER_TITLE
    f"({au10_ti10_ctNC} & {match(VOLUME, YEAR)} & {non_contradicting(NUMBER, PAGES, DOI, ABSTRACT)})",
    f"({au09_ti09_ctXX} & {match(PAGES, DOI)} & {non_contradicting(VOLUME, NUMBER, ABSTRACT)} & {YEAR} > 0.9)",
    f"({au09_ti09_ctXX} & ({match(NUMBER)} & {non_contradicting(PAGES)} | {non_contradicting(NUMBER)} & {match(PAGES)}) & {non_contradicting(VOLUME, YEAR, DOI, ABSTRACT)})",
    f"({au09_ti09_ctXX} & {match(VOLUME, PAGES)})",
    f"({au09_ti09_ctXX} & {match(PAGES, YEAR)} & {non_contradicting(VOLUME, NUMBER, DOI)})",
    # no TITLE
    f"({au10_tiXX_ct10} & {match(VOLUME, NUMBER, PAGES, YEAR)} & {non_contradicting(DOI)} & ({ABSTRACT} > 0.95 | {non_contradicting(ABSTRACT)}))",  # typically for number-mismatches in title
]

non_duplicate_conditions = [
    f"({mismatch(YEAR)} & ~({match(VOLUME)} | {match(NUMBER)} | {match(PAGES)} | {match(DOI)} | {match(CONTAINER_TITLE)}))",
    f'({mismatch(TITLE)} & ({PAGE_RANGES_ADJACENT} == "adjacent" | {PAGE_RANGES_ADJACENT} == "non_overlapping"))',
    f"(~(doi_1 == '' | doi_2 == '') & {DOI} < 0.8 & ~({non_contradicting(AUTHOR, TITLE, YEAR, CONTAINER_TITLE, VOLUME, NUMBER, PAGES)}))",
    f"({mismatch(VOLUME, NUMBER, PAGES)})",
    # Editorials: minor differences in volume/number/pages can be meaningful
    f'(title_1.str.contains("editor") & title_1.str.len() < 60 & ( {mismatch(VOLUME)} | {mismatch(NUMBER)} | {mismatch(PAGES)}))',
]



#! /usr/bin/env python
"""Module to check maybe cases"""
from datetime import datetime
from pathlib import Path

import pandas as pd
from rapidfuzz import fuzz

import bib_dedupe.cluster
from bib_dedupe import verbose_print
from bib_dedupe.constants.fields import ABSTRACT
from bib_dedupe.constants.fields import AUTHOR
from bib_dedupe.constants.fields import CLUSTER_ID
from bib_dedupe.constants.fields import CONTAINER_TITLE
from bib_dedupe.constants.fields import DOI
from bib_dedupe.constants.fields import DUPLICATE
from bib_dedupe.constants.fields import DUPLICATE_LABEL
from bib_dedupe.constants.fields import ID
from bib_dedupe.constants.fields import MAYBE
from bib_dedupe.constants.fields import NUMBER
from bib_dedupe.constants.fields import PAGES
from bib_dedupe.constants.fields import TITLE
from bib_dedupe.constants.fields import VOLUME
from bib_dedupe.constants.fields import YEAR

MAYBE_CASES_FILEPATH = "maybe_cases.csv"
SIMILARITY_SCORE = "similarity_score"
FIELDS_TO_EXPORT = [
    SIMILARITY_SCORE,
    DUPLICATE_LABEL,
    CLUSTER_ID,
    AUTHOR,
    TITLE,
    CONTAINER_TITLE,
    YEAR,
    VOLUME,
    NUMBER,
    PAGES,
    DOI,
    ABSTRACT,
]


def __calculate_similarity(
    author1: str, title1: str, author2: str, title2: str
) -> float:
    concatenated1 = author1 + title1
    concatenated2 = author2 + title2
    return round(fuzz.ratio(concatenated1, concatenated2) / 100, 2)


def export_maybe(
    matched_df: pd.DataFrame,
    records_df: pd.DataFrame,
) -> None:
    """
    This method is used to check maybe cases for deduplication.

    Parameters:
    matched_df (pd.DataFrame): The dataframe containing the matched pairs.
    records_df (pd.DataFrame): The dataframe containing the records.
    """

    duplicate_id_clusters = bib_dedupe.cluster.get_connected_components(
        matched_df, label=DUPLICATE
    )

    # to avoid multiple maybe-pairs,
    # consider only one link between clusters (the first ID of duplicate_id_clusters)
    duplicate_id_dict = {
        i: duplicate_id_cluster[0]
        for duplicate_id_cluster in duplicate_id_clusters
        for i in duplicate_id_cluster
        if i != duplicate_id_cluster[0]
    }
    maybe_df = matched_df.loc[matched_df[DUPLICATE_LABEL] == MAYBE].copy()
    maybe_df.loc[:, f"{ID}_1"] = (
        maybe_df[f"{ID}_1"].map(duplicate_id_dict).fillna(maybe_df[f"{ID}_1"])
    )
    maybe_df.loc[:, f"{ID}_2"] = (
        maybe_df[f"{ID}_2"].map(duplicate_id_dict).fillna(maybe_df[f"{ID}_2"])
    )

    maybe_df = maybe_df.drop_duplicates(subset=[f"{ID}_1", f"{ID}_2"])

    maybe_id_pairs = maybe_df[[f"{ID}_1", f"{ID}_2"]].values.tolist()
    maybe_id_pairs = [pair for pair in maybe_id_pairs if pair[0] != pair[1]]

    maybe_cases_df = pd.DataFrame()
    maybe_cases_df.loc[:, ID] = [id for sublist in maybe_id_pairs for id in sublist]

    maybe_cases_df.loc[:, CLUSTER_ID] = [
        i for i, sublist in enumerate(maybe_id_pairs) for _ in sublist
    ]

    if maybe_cases_df.empty:
        return

    maybe_df = pd.merge(
        maybe_cases_df,
        records_df,
        left_on=ID,
        right_on=ID,
        how="inner",
    )

    similarity_scores = maybe_df.groupby(CLUSTER_ID).apply(
        lambda group: __calculate_similarity(
            group[AUTHOR].iloc[0],
            group[TITLE].iloc[0],
            group[AUTHOR].iloc[1],
            group[TITLE].iloc[1],
        )
    )
    maybe_df = maybe_df.merge(
        similarity_scores.rename(SIMILARITY_SCORE),
        left_on=CLUSTER_ID,
        right_index=True,
    )

    maybe_df = maybe_df.sort_values(
        [SIMILARITY_SCORE, CLUSTER_ID], ascending=[False, True]
    ).reset_index(drop=True)
    maybe_df.insert(0, DUPLICATE_LABEL, MAYBE)

    # Reorder the columns
    maybe_df = maybe_df[
        FIELDS_TO_EXPORT
        + [col for col in maybe_df.columns if col not in FIELDS_TO_EXPORT]
    ]

    file_path = Path(MAYBE_CASES_FILEPATH)
    if file_path.exists():
        file_path.rename(
            file_path.stem
            + "_"
            + datetime.now().strftime("%Y%m%d%H%M%S")
            + file_path.suffix
        )

    maybe_df.to_csv(MAYBE_CASES_FILEPATH, index=False)

    verbose_print.print(
        """Please follow the steps below to replace the 'duplicate_label' column
        with 'duplicate' if it is a duplicate:
        1. Open the 'maybe_cases.csv' file.
        2. Replace the 'duplicate_label' column values with 'duplicate' where necessary.
        3. Save the changes to 'maybe_cases.csv'."""
    )


def import_maybe(
    matched_df: pd.DataFrame,
) -> pd.DataFrame:
    """
    This method is used to import decisions for maybe cases.

    Parameters:
    matched_df (pd.DataFrame): The dataframe containing the matches.

    Returns:
    pd.DataFrame: The dataframe containing the updated matches.
    """

    if not Path(MAYBE_CASES_FILEPATH).is_file():
        verbose_print.print(f"File {MAYBE_CASES_FILEPATH} does not exist")
        return matched_df

    maybe_cases_df = pd.read_csv(MAYBE_CASES_FILEPATH)

    duplicate_cases = (
        maybe_cases_df[maybe_cases_df[DUPLICATE_LABEL] == DUPLICATE]
        .groupby(CLUSTER_ID)
        .apply(lambda group: set(group[ID].values.flatten().astype(str)))
        .values.tolist()
    )

    verbose_print.print(f"Number of cases labeled as duplicate: {len(duplicate_cases)}")

    # Replace the 'duplicate_label' column in matched_df dataframe with 'duplicate' for these cases
    if len(duplicate_cases):
        matched_df.loc[
            matched_df.apply(
                lambda row: {row[f"{ID}_1"], row[f"{ID}_2"]} in duplicate_cases, axis=1
            ),
            DUPLICATE_LABEL,
        ] = DUPLICATE

    matched_df = matched_df[matched_df[DUPLICATE_LABEL] != MAYBE]

    return matched_df



#! /usr/bin/env python
"""Utility to export and evaluate dedupe benchmarks."""
from __future__ import annotations

import time
import typing
from datetime import datetime
from itertools import combinations
from pathlib import Path
from typing import Optional

import pandas as pd
from tqdm import tqdm

import bib_dedupe.cluster
import bib_dedupe.util
from bib_dedupe.constants.fields import AUTHOR_FIRST
from bib_dedupe.constants.fields import CONTAINER_TITLE_SHORT
from bib_dedupe.constants.fields import DUPLICATE_LABEL
from bib_dedupe.constants.fields import ID
from bib_dedupe.constants.fields import MAYBE
from bib_dedupe.constants.fields import ORIGIN
from bib_dedupe.constants.fields import TITLE_SHORT

TP = "TP"
FP = "FP"
TN = "TN"
FN = "FN"
PRECISION = "precision"
SENSITIVITY = "sensitivity"
SPECIFICITY = "specificity"
FP_RATE = "false_positive_rate"
F1 = "f1"
RUNTIME = "runtime"

CASE = "case"


class DedupeBenchmarker:
    """Dedupe benchmarker"""

    # pylint: disable=too-many-instance-attributes

    true_merged_ids: list
    records_df: pd.DataFrame

    def __init__(
        self,
        *,
        benchmark_path: Optional[Path] = None,
        regenerate_benchmark_from_history: bool = False,
        colrev_project_path: Optional[Path] = None,
    ) -> None:
        if benchmark_path is None:
            benchmark_path = Path.cwd()
        self.benchmark_path = Path(benchmark_path).resolve()
        if colrev_project_path is None:
            self.colrev_project_path = benchmark_path
        else:
            self.colrev_project_path = colrev_project_path

        self.records_pre_merged_path = Path(
            self.benchmark_path, "records_pre_merged.csv"
        )
        self.merged_record_ids_path = Path(self.benchmark_path, "merged_record_ids.csv")
        if regenerate_benchmark_from_history:
            self.__get_dedupe_benchmark()
        else:
            self.__load_data()

    def __load_data(self) -> None:
        # print("after", self.true_merged_ids)

        if self.records_pre_merged_path.is_file():
            records_df = pd.read_csv(str(self.records_pre_merged_path))
            self.records_df = records_df
        else:
            part1_path = self.records_pre_merged_path.with_name(
                self.records_pre_merged_path.stem + "_part1.csv"
            )
            part2_path = self.records_pre_merged_path.with_name(
                self.records_pre_merged_path.stem + "_part2.csv"
            )

            if part1_path.is_file() and part2_path.is_file():
                part1_df = pd.read_csv(str(part1_path))
                self.records_df = part1_df
                part2_df = pd.read_csv(str(part2_path))
                self.records_df = pd.concat(
                    [self.records_df, part2_df], ignore_index=True
                )

        true_merged_ids_df = pd.read_csv(str(self.merged_record_ids_path))
        self.true_merged_ids = true_merged_ids_df["merged_ids"].str.split(";").tolist()

        # Validate
        non_unique_lists = []
        for merged_id_list in self.true_merged_ids:
            if len(merged_id_list) != len(set(merged_id_list)):
                non_unique_lists.append(merged_id_list)

        for non_unique_list in non_unique_lists:
            print(f"Duplicate strings found in: {non_unique_list}")

        if non_unique_lists:
            raise ValueError(
                "Each list in self.true_merged_ids should only contain unique strings"
            )

        all_ids_in_true_merged_ids = [
            id for sublist in self.true_merged_ids for id in sublist
        ]
        ids_not_in_records_df = set(all_ids_in_true_merged_ids) - set(
            self.records_df[ID].tolist()
        )
        if ids_not_in_records_df:
            print(f"IDs not found in records_df: {ids_not_in_records_df}")
            raise ValueError(
                "Not all ids in self.true_merged_ids are in self.records_df column ID"
            )

        # Validate: Ensure that there is no overlap between sets of self.true_merged_ids
        flattened_ids = [id for sublist in self.true_merged_ids for id in sublist]
        if len(flattened_ids) != len(set(flattened_ids)):
            from collections import Counter

            counter = Counter(flattened_ids)
            problem_ids = [id for id, count in counter.items() if count > 1]
            print(f"Problem IDs: {problem_ids}")
            raise ValueError("Overlap found between sets in self.true_merged_ids")

    def get_records_for_dedupe(self) -> pd.DataFrame:
        """
        Get (pre-processed) records for dedupe

        Returns:
            pd.DataFrame: Pre-processed records for dedupe
        """

        return self.records_df

    # pylint: disable=too-many-locals
    def __get_dedupe_benchmark(self) -> None:
        """Get benchmark for dedupe"""
        import colrev.review_manager
        import colrev.record

        STATUS = "colrev_status"

        def merged(record: dict) -> bool:
            return len([o for o in record[ORIGIN] if not o.startswith("md_")]) != 1

        self.review_manager = colrev.review_manager.ReviewManager(
            path_str=str(self.colrev_project_path), force_mode=True
        )
        self.dedupe_operation = self.review_manager.get_dedupe_operation()

        records = self.review_manager.dataset.load_records_dict()
        for record in records.values():
            if STATUS not in record:
                record[STATUS] = colrev.record.RecordState.md_processed
                print("setting missing status")

        # Select md-processed records (discard recently added/non-deduped ones)
        records = {
            r[ID]: r
            for r in records.values()
            if r[STATUS]
            in colrev.record.RecordState.get_post_x_states(
                state=colrev.record.RecordState.md_processed
            )
        }
        # Drop origins starting with md_... (not relevant for dedupe)
        for record_dict in records.values():
            record_dict[ORIGIN] = [
                o for o in record_dict[ORIGIN] if not o.startswith("md_")
            ]

        records_pre_merged_list = [r for r in records.values() if not merged(r)]
        records_merged_origins = [
            o
            for r in records.values()
            if merged(r)
            for o in r[ORIGIN]
            if not o.startswith("md_")
        ]

        nr_commits = self.review_manager.dataset.get_repo().git.rev_list(
            "--count", "HEAD"
        )
        for hist_recs in tqdm(
            self.review_manager.dataset.load_records_from_history(),
            total=int(nr_commits),
        ):
            if len(records_merged_origins) == 0:
                break

            try:
                for hist_record_dict in hist_recs.values():
                    if any(
                        o in hist_record_dict[ORIGIN] for o in records_merged_origins
                    ) and not merged(hist_record_dict):
                        # only consider post-md_prepared (non-merged) records
                        if hist_record_dict[
                            STATUS
                        ] in colrev.record.RecordState.get_post_x_states(
                            state=colrev.record.RecordState.md_prepared
                        ):
                            hist_record_dict[ORIGIN] = [
                                o
                                for o in hist_record_dict[ORIGIN]
                                if not o.startswith("md_")
                            ]
                            records_pre_merged_list.append(hist_record_dict)
                            records_merged_origins.remove(
                                [
                                    o
                                    for o in hist_record_dict[ORIGIN]
                                    if not o.startswith("md_")
                                ][0]
                            )

            except KeyError:
                break

        records_pre_merged = {r[ID]: r for r in records_pre_merged_list}

        # drop missing from records
        # (can only evaluate record/origins that are available in records_pre_merged)
        pre_merged_orgs = {o for r in records_pre_merged.values() for o in r[ORIGIN]}
        for record_dict in records.values():
            record_dict[ORIGIN] = [
                o
                for o in record_dict[ORIGIN]
                if o in pre_merged_orgs and o not in records_merged_origins
            ]
        records = {r[ID]: r for r in records.values() if len(r[ORIGIN]) > 0}

        assert {o for x in records_pre_merged.values() for o in x[ORIGIN]} == {
            o for x in records.values() for o in x[ORIGIN]
        }
        # [x for x in o_rec if x not in o_pre]

        records_pre_merged_df = pd.DataFrame.from_dict(
            records_pre_merged, orient="index"
        )
        self.records_pre_merged_df = records_pre_merged_df

        records_df = pd.DataFrame.from_dict(records, orient="index")
        self.records_df = records_df

        merged_record_ids = []
        for row in list(records_df.to_dict(orient="records")):
            if len(row[ID]) > 1:
                merged_record_ids.append(row[ID])

        self.true_merged_ids = merged_record_ids
        merged_record_ids_df = pd.DataFrame({"merged_ids": self.true_merged_ids})
        records_pre_merged_df.to_csv(str(self.records_pre_merged_path), index=False)
        merged_record_ids_df.to_csv(str(self.merged_record_ids_path), index=False)

    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    def compare(
        self,
        *,
        blocked_df: pd.DataFrame,
        predicted: list,
    ) -> dict:
        """Compare the predicted matches and blocked pairs to the ground truth."""

        ground_truth_pairs = set()
        for item in self.true_merged_ids:
            for combination in combinations(item, 2):
                ground_truth_pairs.add(";".join(sorted(combination)))

        blocked = blocked_df.apply(lambda row: [row[f"{ID}_1"], row[f"{ID}_2"]], axis=1)

        blocked_pairs = set()
        for item in blocked:
            for combination in combinations(item, 2):
                blocked_pairs.add(";".join(sorted(combination)))

        predicted_pairs = set()
        for item in predicted:
            for combination in combinations(item, 2):
                predicted_pairs.add(";".join(sorted(combination)))

        blocks = {TP: 0, FP: 0, TN: 0, FN: 0}
        blocks_fn_list = []
        matches = {TP: 0, FP: 0, TN: 0, FN: 0}
        matches_fp_list = []
        matches_fn_list = []

        all_ids = self.records_df[ID].tolist()

        # Note: the following takes long:
        for combination in combinations(all_ids, 2):
            pair = ";".join(sorted(combination))

            if pair in blocked_pairs:
                if pair in ground_truth_pairs:
                    blocks[TP] += 1
                else:
                    blocks[FP] += 1
                    # Don't need a list here.
            else:
                if pair in ground_truth_pairs:
                    blocks[FN] += 1
                    blocks_fn_list.append(combination)
                else:
                    blocks[TN] += 1

            if pair in predicted_pairs:
                if pair in ground_truth_pairs:
                    matches[TP] += 1
                else:
                    matches[FP] += 1
                    matches_fp_list.append(combination)
            else:
                if pair in ground_truth_pairs:
                    matches[FN] += 1
                    matches_fn_list.append(combination)
                else:
                    matches[TN] += 1

        return {
            "matches": matches,
            "blocks_FN_list": blocks_fn_list,
            "matches_FP_list": matches_fp_list,
            "matches_FN_list": matches_fn_list,
            "blocks": blocks,
        }

    def get_runtime(self, timestamp: datetime) -> str:
        """
        Calculate the runtime.

        Args:
            timestamp (datetime.datetime): The timestamp when the deduplication was started.

        Returns:
            str: The runtime in the format "hours:minutes:seconds".
        """
        start_time = datetime.now()
        time_diff = start_time - timestamp
        runtime_in_minutes = time_diff.total_seconds() / 60

        runtime_in_hours = runtime_in_minutes / 60
        hours = int(runtime_in_hours)
        minutes = (runtime_in_hours * 60) % 60
        seconds = (runtime_in_hours * 3600) % 60
        runtime_string = f"{hours}:{int(minutes):02d}:{int(seconds):02d}"
        print(f"Runtime: {runtime_string}")
        return runtime_string

    def compare_dedupe_id(
        self,
        *,
        records_df: pd.DataFrame,
        merged_df: pd.DataFrame,
        timestamp: datetime,
    ) -> pd.DataFrame:
        """
        Compare dedupe IDs and calculate evaluation metrics.

        Args:
            records_df (pd.DataFrame): DataFrame containing the original records.
            merged_df (pd.DataFrame): DataFrame containing the merged records.
            timestamp (datetime.datetime): The timestamp when the deduplication was started.

        Returns:
            pd.DataFrame: DataFrame containing the evaluation metrics.
        """

        results: typing.Dict[str, typing.Any] = {
            TP: 0,
            FP: 0,
            FN: 0,
            TN: 0,
            RUNTIME: self.get_runtime(timestamp),
        }

        all_ids = records_df[ID].tolist()

        true_non_duplicate_ids = [
            x for x in all_ids if not any(x in y for y in self.true_merged_ids)
        ]

        # Assume that IDs are not changed (merged IDs are not available)
        # For each ID-set, **exactly one** must be in the merged_df

        for true_non_duplicate_id in true_non_duplicate_ids:
            if true_non_duplicate_id in merged_df[ID].tolist():
                results[TN] += 1
            else:
                results[FP] += 1

        for true_merged_id_set in self.true_merged_ids:
            nr_in_merged_df = merged_df[merged_df[ID].isin(true_merged_id_set)].shape[0]
            # One would always be required to be a non-duplicate (true value:negative)
            # All that are removed are true positive, all that are not removed are false negatives
            if nr_in_merged_df == 0:
                results[FP] += 1
                results[TP] += len(true_merged_id_set) - 1
            elif nr_in_merged_df >= 1:
                results[TN] += 1
                results[FN] += nr_in_merged_df - 1
                results[TP] += len(true_merged_id_set) - nr_in_merged_df

        results[FP_RATE] = results[FP] / (results[FP] + results[TN])

        results[SPECIFICITY] = results[TN] / (results[TN] + results[FP])
        results[SENSITIVITY] = results[TP] / (results[TP] + results[FN])
        if (results[TP] + results[FP]) > 0:
            results[PRECISION] = results[TP] / (results[TP] + results[FP])
        else:
            results[PRECISION] = 0.0
        if (results[PRECISION] + results[SENSITIVITY]) > 0:
            results[F1] = (
                2
                * (results[PRECISION] * results[SENSITIVITY])
                / (results[PRECISION] + results[SENSITIVITY])
            )
        else:
            results[F1] = 0.0

        results["dataset"] = Path(self.benchmark_path).name

        return results

    def export_cases(
        self,
        *,
        prepared_records_df: pd.DataFrame,
        blocked_df: pd.DataFrame,
        matched_df: pd.DataFrame,
    ) -> None:
        """Get the cases for results

        records_df = [ID, title, author, ...]
        blocked_df = ...
        results = {"blocks_FN_list", "matches_FP_list", "matches_FN_list"}
        """

        maybe_cases_df = matched_df[matched_df[DUPLICATE_LABEL] == MAYBE].copy()

        columns_to_drop = [
            TITLE_SHORT,
            AUTHOR_FIRST,
            CONTAINER_TITLE_SHORT,
            ORIGIN,
        ]
        prepared_records_df = prepared_records_df.drop(
            columns=[
                col for col in columns_to_drop if col in prepared_records_df.columns
            ]
        )

        maybe_cases_df.loc[:, CASE] = (
            maybe_cases_df[f"{ID}_1"] + ";" + maybe_cases_df[f"{ID}_2"]
        )
        maybe_df_copy = maybe_cases_df.copy()
        maybe_df_1 = pd.merge(
            maybe_df_copy,
            prepared_records_df,
            left_on=f"{ID}_2",
            right_on=ID,
            how="inner",
        )
        maybe_df_2 = pd.merge(
            maybe_cases_df,
            prepared_records_df,
            left_on=f"{ID}_1",
            right_on=ID,
            how="inner",
        )

        maybe_df = pd.concat([maybe_df_1, maybe_df_2])
        maybe_df = maybe_df.sort_values(by=CASE)
        maybe_df = maybe_df.drop(columns=[f"{ID}_1", f"{ID}_2", DUPLICATE_LABEL])

        maybe_df.to_csv(self.benchmark_path / "maybe_pairs.csv", index=False)

        print("Export started at", datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
        start_time = time.time()

        duplicate_id_sets = bib_dedupe.cluster.get_connected_components(matched_df)

        results = self.compare(
            blocked_df=blocked_df,
            predicted=duplicate_id_sets,
        )

        for list_name in [
            "blocks_FN_list",
            "matches_FP_list",
            "matches_FN_list",
        ]:
            id_pairs = results[list_name]

            id_pairs_cases_df = pd.DataFrame(id_pairs, columns=[f"{ID}_1", f"{ID}_2"])
            id_pairs_cases_df.loc[:, CASE] = (
                id_pairs_cases_df[f"{ID}_1"] + ";" + id_pairs_cases_df[f"{ID}_2"]
            )

            id_pairs_cases_df_copy = id_pairs_cases_df.copy()
            id_pairs_df_1 = pd.merge(
                id_pairs_cases_df_copy,
                prepared_records_df,
                left_on=f"{ID}_2",
                right_on=ID,
                how="inner",
            )
            id_pairs_df_2 = pd.merge(
                id_pairs_cases_df,
                prepared_records_df,
                left_on=f"{ID}_1",
                right_on=ID,
                how="inner",
            )

            cases_df = pd.concat([id_pairs_df_1, id_pairs_df_2])
            cases_df = cases_df.sort_values(by=CASE)
            cases_df = cases_df.drop(columns=[f"{ID}_1", f"{ID}_2"])

            if not cases_df.empty:
                cases_df = cases_df[
                    [CASE, ID]
                    + [col for col in cases_df.columns if col not in [CASE, ID]]
                ]
                cases_df = cases_df.sort_values(by=[CASE, ID])

                if cases_df.empty:
                    continue

            cases_df.to_csv(self.benchmark_path / f"{list_name}.csv", index=False)

            ignored_file = self.benchmark_path / f"{list_name}_ignored.csv"
            if not ignored_file.is_file():
                continue
            try:
                ignored_df = pd.read_csv(ignored_file)
                cases_df = cases_df[~cases_df[CASE].isin(ignored_df[CASE])]
                cases_df.to_csv(
                    self.benchmark_path / f"{list_name}_remaining.csv", index=False
                )
            except KeyError as exc:
                print(exc)

        end_time = time.time()
        print(f"Export completed after: {end_time - start_time:.2f} seconds")



#! /usr/bin/env python
"""Cluster for dedupe"""
from collections import defaultdict
from itertools import combinations
from typing import List
from typing import Set

import pandas as pd

from bib_dedupe.constants.fields import DUPLICATE_LABEL


def get_adjacency_list(duplicates_df: pd.DataFrame) -> dict:
    """
    Generate an adjacency list from a DataFrame of duplicate pairs.

    Args:
        duplicates_df (pd.DataFrame): A DataFrame where each row represents a pair of duplicate IDs.

    Returns:
        dict: An adjacency list. Each key is an ID and each value is a list
        of IDs that are duplicates of the key.
    """
    id_sets = duplicates_df[["ID_1", "ID_2"]].values.tolist()

    adjacency_list = defaultdict(list)

    for id_set in id_sets:
        for combination in combinations(id_set, 2):
            adjacency_list[combination[0]].append(combination[1])
            adjacency_list[combination[1]].append(combination[0])
    return adjacency_list


def depth_first_search(
    node: str,
    adjacency_list: dict,
    visited: dict,
    component: list,
    entity_set_map: dict,
    component_sets: set,
) -> None:
    """
    Perform a depth-first search on an adjacency list starting from a given node.

    Args:
        node (str): The node to start the search from.
        adjacency_list (dict): The adjacency list to perform the search on.
        visited (dict): A dictionary where each key is a node and
        each value is a boolean indicating if the node has been visited.
        component (list): A list to store the nodes visited in the search.
        entity_set_map (dict): A mapping of entity IDs to their sets.
        component_sets (set): A set to store the sets that have been added to the component.
    """

    node_set = entity_set_map[node]

    if node_set in component_sets:
        return

    visited[node] = True
    component.append(node)
    if node_set:
        component_sets.add(node_set)

    for neighbor in adjacency_list[node]:
        if not visited[neighbor]:
            depth_first_search(
                neighbor,
                adjacency_list,
                visited,
                component,
                entity_set_map,
                component_sets,
            )


def get_connected_components(
    duplicates_df: pd.DataFrame,
    label: str = "duplicate",
) -> list:
    """
    Find the connected components in an adjacency list represented by a DataFrame of duplicate pairs,
    while ensuring no component contains more than one entity from the same set.

    Args:
        duplicates_df (pd.DataFrame): A DataFrame where each row represents a pair of duplicate IDs.
        label (str, optional): The label to filter the DataFrame by. Defaults to "duplicate".

    Returns:
        list: A list of connected components. Each component is a list of IDs
             that are all duplicates of each other, with no two entities from the same set.
    """

    if duplicates_df.empty:
        return []

    duplicates_df = duplicates_df[duplicates_df[DUPLICATE_LABEL] == label]

    adjacency_list = get_adjacency_list(duplicates_df)

    entity_set_map = {}

    for _, row in duplicates_df.iterrows():
        entity_set_map[row["ID_1"]] = row["search_set_1"]
        entity_set_map[row["ID_2"]] = row["search_set_2"]

    visited = {node: False for node in adjacency_list}
    components = []

    for node in adjacency_list:
        if not visited[node]:
            component: List[str] = []
            component_sets: Set[str] = set()
            depth_first_search(
                node, adjacency_list, visited, component, entity_set_map, component_sets
            )
            components.append(sorted(component))

    return components



#!/usr/bin/env python
"""ENTRYTYPE constants for bib-dedupe"""
# pylint: disable=too-few-public-methods


ARTICLE = "article"
INPROCEEDINGS = "inproceedings"
BOOK = "book"
INBOOK = "inbook"
PROCEEDINGS = "proceedings"

INCOLLECTION = "incollection"
PHDTHESIS = "phdthesis"
THESIS = "thesis"
MASTERSTHESIS = "masterthesis"
BACHELORTHESIS = "bachelorthesis"
TECHREPORT = "techreport"
UNPUBLISHED = "unpublished"
MISC = "misc"
SOFTWARE = "software"
ONLINE = "online"
CONFERENCE = "conference"



#!/usr/bin/env python
"""Field constants for bib-dedupe"""
# pylint: disable=too-few-public-methods


ID = "ID"
ENTRYTYPE = "ENTRYTYPE"
DOI = "doi"
URL = "url"
# TBD: no LINK field?
ISSN = "issn"
ISBN = "isbn"
FULLTEXT = "fulltext"
ABSTRACT = "abstract"
KEYWORDS = "keywords"
CITED_BY = "cited_by"
FILE = "file"
INSTITUTION = "institution"
MONTH = "month"
SERIES = "series"
SCHOOL = "school"
LANGUAGE = "language"

ORIGIN = "origin"
SEARCH_SET = "search_set"

TITLE = "title"
AUTHOR = "author"
YEAR = "year"
JOURNAL = "journal"
BOOKTITLE = "booktitle"
CHAPTER = "chapter"
PUBLISHER = "publisher"
VOLUME = "volume"
NUMBER = "number"
PAGES = "pages"
EDITOR = "editor"
EDITION = "edition"
ADDRESS = "address"

CONTAINER_TITLE = "container_title"

# For blocking
AUTHOR_FIRST = "author_first"
TITLE_SHORT = "title_short"
CONTAINER_TITLE_SHORT = "container_title_short"

# For similarity
PAGE_RANGES_ADJACENT = "page_ranges_adjacent"

DUPLICATE_LABEL = "duplicate_label"
CLUSTER_ID = "cluster_ID"

DUPLICATE = "duplicate"
MAYBE = "maybe"



#! /usr/bin/env python
"""Preparation of abstract field"""
import re

import numpy as np


def prep_abstract(abstract_array: np.array) -> np.array:
    abstract_array = np.array(
        [re.sub(r"<.*?>", " ", abstract.lower()) for abstract in abstract_array]
    )

    abstract_array = np.array(
        [
            re.sub(r"^aims\s|^objectives|^background", "", abstract)
            for abstract in abstract_array
        ]
    )

    abstract_array = np.array(
        [
            abstract[: abstract.rfind(". copyright")]
            if ". copyright" in abstract[-300:]
            else abstract[: abstract.rfind("©")]
            if "©" in abstract[-200:]
            else re.sub(r"(\s*\d{4}\s*)?the authors[.?]$", "", abstract)
            if "the authors" in abstract[-100:]
            else abstract[: abstract.rfind("springer-verlag")]
            if "springer-verlag" in abstract[-100:]
            else re.sub(r"\s*\d{4}.*$", "", abstract)
            if re.search(r"\.\s*\d{4}.*$", abstract)
            else re.sub(r" \(c\) \d{4}.*\.$", "", abstract)
            if re.search(r"\. \(c\) \d{4}.*\.$", abstract)
            else re.sub(r"\.\(abstract truncated at 400 words\)$", "", abstract)
            if ".(abstract truncated at 400 words)" in abstract[-80:]
            else abstract
            for abstract in abstract_array
        ]
    )

    # Remove "abstract " at the beginning
    abstract_array = np.array(
        [re.sub(r"^abstract ", "", abstract) for abstract in abstract_array]
    )

    # Remove trailing date
    abstract_array = np.array(
        [re.sub(r"\s*\(\d{4}\)$", "", abstract) for abstract in abstract_array]
    )

    abstract_array = np.array(
        [re.sub(r"[^A-Za-z0-9 .,]", "", abstract) for abstract in abstract_array]
    )
    abstract_array = np.array(
        [re.sub(r"\s+", " ", abstract) for abstract in abstract_array]
    )
    return np.array(
        [
            "" if abstract == "nan" else abstract.lower().rstrip(" .").lstrip(" .")
            for abstract in abstract_array
        ]
    )



#! /usr/bin/env python
"""Preparation of year field"""
import re

import numpy as np


def prep_year(year_array: np.array) -> np.array:
    def process_year(value: str) -> str:
        try:
            match = re.match(r"^(\d{4})-\d{2}-\d{2}$", value)
            if match:
                value = match.group(1)

            int_value = int(float(value))

            if not 1900 < int_value < 2100:
                return ""

        except ValueError:
            return ""
        return str(int_value)

    return np.array([process_year(year) for year in year_array])



#! /usr/bin/env python
"""BibDedupe utility"""
from __future__ import annotations

import pprint


class VerbosePrint:
    """A class to handle verbose printing with different verbosity levels."""

    def __init__(self, verbosity_level: int):
        """Initialize the VerbosePrint class with a specific verbosity level."""
        self.verbosity_level = verbosity_level
        self.p_printer = pprint.PrettyPrinter(indent=4, width=140, compact=False)

    def print(self, message: str = "", *, level: int = 1) -> None:
        """Print a message if the specified level is less than or equal to the current verbosity level."""
        if level <= self.verbosity_level:
            print(message)

    def pretty_print(self, input_dict: dict) -> None:
        """Pretty print a dictionary."""
        self.p_printer.pprint(input_dict)



#! /usr/bin/env python
"""Debug for dedupe"""
from pathlib import Path

import pandas as pd

from bib_dedupe import verbose_print
from bib_dedupe.bib_dedupe import block
from bib_dedupe.bib_dedupe import match
from bib_dedupe.bib_dedupe import prep
from bib_dedupe.dedupe_benchmark import DedupeBenchmarker


def debug() -> None:
    try:
        df_blocks = pd.read_csv("blocks_FN_list.csv")
    except (pd.errors.EmptyDataError, FileNotFoundError):
        df_blocks = pd.DataFrame()

    try:
        df_matches = pd.read_csv("matches_FN_list.csv")
    except (pd.errors.EmptyDataError, FileNotFoundError):
        df_matches = pd.DataFrame()

    try:
        df_matches_fp = pd.read_csv("matches_FP_list.csv")
    except (pd.errors.EmptyDataError, FileNotFoundError):
        df_matches_fp = pd.DataFrame()

    dedupe_benchmark = DedupeBenchmarker(benchmark_path=Path.cwd())

    records_df = dedupe_benchmark.get_records_for_dedupe()
    records_df = prep(records_df)

    while True:
        id_pair = input("id_pair:")

        try:
            if "case" in df_blocks.columns and id_pair in df_blocks["case"].values:
                print(f"ID pair {id_pair} found in blocks_FN_list.csv")
            if "case" in df_matches.columns and id_pair in df_matches["case"].values:
                print(f"ID pair {id_pair} found in matches_FN_list.csv")
            if (
                "case" in df_matches_fp.columns
                and id_pair in df_matches_fp["case"].values
            ):
                print(f"ID pair {id_pair} found in matches_FP_list.csv")

            id1, id2 = id_pair.split(";")
            selected_prepared_records_df = records_df[
                records_df["ID"].apply(lambda x: id1 == x or id2 == x)
            ]

            if selected_prepared_records_df.empty:
                print("selected_prepared_records_df is empty. Continuing...")
                continue

            actual_blocked_df = block(
                records_df=selected_prepared_records_df, verbosity_level=2
            )
            matches = match(actual_blocked_df, verbosity_level=2)
            verbose_print.p_printer.pprint(matches)
        except ValueError as exc:
            print(exc)



import click

import bib_dedupe.debug


@click.group()
def main() -> None:
    """A simple CLI for bib_dedupe."""
    pass


@main.command()
@click.pass_context
def debug(ctx: click.Context) -> None:
    bib_dedupe.debug.debug()


if __name__ == "__main__":
    main()



#! /usr/bin/env python
"""Debug for dedupe"""
import pandas as pd

from bib_dedupe import verbose_print
from bib_dedupe.constants.fields import AUTHOR
from bib_dedupe.constants.fields import JOURNAL
from bib_dedupe.constants.fields import ORIGIN
from bib_dedupe.constants.fields import PAGES
from bib_dedupe.constants.fields import TITLE
from bib_dedupe.constants.fields import YEAR


def percent_upper_chars(input_string: str) -> float:
    """Get the percentage of upper-case characters in a string"""
    if len(input_string) == 0:
        return 0
    return sum(map(str.isupper, input_string)) / len(input_string)


def default_merge_function_title(titles: list) -> str:
    """Default merge function for title field"""

    if len([title for title in titles if not (pd.isnull(title) or title == "")]) <= 1:
        return titles[0]

    best_title = titles[0]

    # Note : avoid switching titles
    for title in titles[1:]:
        if best_title.replace(" - ", ": ") == title.replace(" - ", ": "):
            return best_title

    best_title_upper = percent_upper_chars(best_title)

    for title in titles[1:]:
        title_upper = percent_upper_chars(title)

        if title[-1] not in ["*", "1", "2"]:
            # Relatively simple rule...
            # catches cases when best_title is all upper or title case
            if best_title_upper > title_upper:
                best_title = title
    return best_title


def default_merge_function_author(authors: list) -> str:
    """Default merge function for author field"""

    if (
        len([author for author in authors if not (pd.isnull(author) or author == "")])
        <= 1
    ):
        return authors[0]

    best_author = authors[0]

    best_author_upper = percent_upper_chars(best_author)

    for author in authors[1:]:
        author_upper = percent_upper_chars(author)

        # Prefer title case (not all-caps)
        if best_author_upper > 0.8 and author_upper <= 0.8:
            best_author = author
    return best_author


def default_merge_function_container_title(journals: list) -> str:
    """Default merge function for container-title field"""

    if (
        len(
            [
                journal
                for journal in journals
                if not (pd.isnull(journal) or journal == "")
            ]
        )
        <= 1
    ):
        return journals[0]

    journals = [
        journal for journal in journals if not (pd.isnull(journal) or journal == "")
    ]

    best_journal = journals[0]

    best_journal_upper = percent_upper_chars(best_journal)

    for journal in journals[1:]:
        journal_upper = percent_upper_chars(journal)

        # Simple heuristic to avoid abbreviations
        if "." in best_journal and "." not in journal:
            best_journal = journal
        # Relatively simple rule...
        # catches cases when best_journal is all upper or title case
        if best_journal_upper > journal_upper:
            best_journal = journal
    return best_journal


def default_merge_function_year(years: list) -> str:
    """Default merge function for year field"""
    # max() to select published version when merging with forthcoming
    years = [str(year) for year in years if not (pd.isnull(year) or year == "")]

    if not any(year.isdigit() for year in years):
        return ""
    return str(max(int(year) for year in years if year.isdigit()))


def default_merge_function_pages(pages: list) -> str:
    """Default merge function for pages field"""

    if len([page for page in pages if not (pd.isnull(page) or page == "")]) <= 1:
        return pages[0]

    pages = [page for page in pages if not (pd.isnull(page) or page == "")]

    best_pages = pages[0]
    for page in pages[1:]:
        if "--" in page and "--" not in best_pages:
            best_pages = page
    return best_pages


def default_merge_function_origin(origins: list) -> str:
    """Default merge function for origin field"""

    unique_origins = set()
    for origin in origins:
        unique_origins.update(origin.split(";"))
    return ";".join(sorted(unique_origins))


DEFAULT_MERGE_FUNCTIONS = {
    ORIGIN: default_merge_function_origin,
    TITLE: default_merge_function_title,
    AUTHOR: default_merge_function_author,
    YEAR: default_merge_function_year,
    JOURNAL: default_merge_function_container_title,
    PAGES: default_merge_function_pages,
}


def merge(
    records_df: pd.DataFrame,
    *,
    duplicate_id_sets: list,
    merge_functions: dict = DEFAULT_MERGE_FUNCTIONS,
    origin_column: str = ORIGIN,
) -> pd.DataFrame:
    """
    This function merges duplicate records in a DataFrame.

    Parameters:
    records_df (pd.DataFrame): The DataFrame containing the records.
    duplicate_id_sets (list): A list of sets, each containing IDs of duplicate records.
    merge_functions (dict, optional): A dictionary mapping column names to functions
    that determine how to merge values in that column. If not provided, default merge
    functions will be used.
    origin_column (str, optional): The name of the column to use as the origin.
    Defaults to ORIGIN.

    Returns:
    pd.DataFrame: The DataFrame with duplicate records merged.
    """

    if not records_df["ID"].is_unique:
        raise ValueError("ID column in records_df must be unique.")

    for duplicate_ids in duplicate_id_sets:
        if not set(duplicate_ids).issubset(set(records_df["ID"].tolist())):
            raise ValueError("Not all duplicate IDs are in the records DataFrame.")

    # Cast all columns in records_df to strings
    records_df = records_df.astype(str)

    if origin_column not in records_df.columns:
        verbose_print.print(f"Add missing origin column ({origin_column})", level=2)
        records_df[origin_column] = records_df["ID"]

    if origin_column not in merge_functions:
        verbose_print.print("Add default_merge_function_origin", level=2)
        merge_functions[origin_column] = DEFAULT_MERGE_FUNCTIONS[ORIGIN]
    if merge_functions == DEFAULT_MERGE_FUNCTIONS:
        verbose_print.print("Use DEFAULT_MERGE_FUNCTIONS", level=2)
    else:
        # Add missing items to merge_function based on DEFAULT_MERGE_FUNCTIONS
        for key, value in DEFAULT_MERGE_FUNCTIONS.items():
            if key not in merge_functions:
                verbose_print.print(
                    f"For {key} add {DEFAULT_MERGE_FUNCTIONS[key].__name__}()", level=2
                )
                merge_functions[key] = value

    non_duplicate_ids = set(records_df["ID"].tolist())
    for duplicate_ids in duplicate_id_sets:
        non_duplicate_ids -= set(duplicate_ids)
        # Apply custom merge functions
        for column, merge_func in merge_functions.items():
            if column in records_df.columns:
                values = records_df.loc[
                    records_df["ID"].isin(duplicate_ids), column
                ].tolist()
                merged_value = merge_func(values)
                verbose_print.print(f"{column}: {values} -> {merged_value}", level=2)
                records_df.loc[
                    records_df["ID"].isin(duplicate_ids), column
                ] = merged_value

    # NOTE: merge_functions must return string. cannot assign object to pandas cell!
    for non_duplicate_id in non_duplicate_ids:
        for column, merge_func in merge_functions.items():
            if column in records_df.columns:
                if column != "nr_intext_citations":
                    continue
                value = records_df.loc[
                    records_df["ID"] == non_duplicate_id, column
                ].tolist()
                merged_value = merge_func(value)
                records_df.loc[
                    records_df["ID"] == non_duplicate_id, column
                ] = merged_value

    to_drop = [o for ol in duplicate_id_sets for o in ol[1:]]

    merged_df = records_df[~records_df["ID"].isin(to_drop)]

    return merged_df



#! /usr/bin/env python
"""Preparation of container_title field"""
import re
from pathlib import Path

import numpy as np
import pandas as pd
from number_parser import parse

from bib_dedupe.constants.entrytypes import ARTICLE
from bib_dedupe.constants.entrytypes import BOOK
from bib_dedupe.constants.entrytypes import INBOOK
from bib_dedupe.constants.entrytypes import INPROCEEDINGS
from bib_dedupe.constants.entrytypes import PROCEEDINGS
from bib_dedupe.constants.fields import BOOKTITLE
from bib_dedupe.constants.fields import CONTAINER_TITLE
from bib_dedupe.constants.fields import ENTRYTYPE
from bib_dedupe.constants.fields import JOURNAL
from bib_dedupe.constants.fields import TITLE

current_dir = Path(__file__).parent
journal_variants_path = current_dir / "journal_variants.csv"
journal_variants = pd.read_csv(journal_variants_path)
JOURNAL_TRANSLATIONS_DICT = dict(
    dict(zip(journal_variants["title_variant"], journal_variants["journal"]))
)
JOURNAL_TRANSLATIONS_DICT = {
    k.lower(): v.lower() for k, v in JOURNAL_TRANSLATIONS_DICT.items()
}


JOURNAL_STOPWORDS = [
    "of",
    "for",
    "the",
    "and",
    "de",
    "d",
    "et",
    "in",
    "i",
    "&",
    "to",
    "on",
    "die",
    "part",
    "annual",
    "und",
    "für",
    "fur",
    "der",
]

JOURNAL_ABBREV = {
    "amer": "am",
    "soci": "soc",
    "expe": "exp",
    "mole": "mol",
    "scie": "sci",
    "brit": "br",
    "bole": "bol",
    "inte": "int",
    "arqu": "arq",
    "polo": "pol",
    "vete": "vet",
    "desi": "des",
    "mede": "med",
    "tera": "ter",
    "huma": "hum",
    "revu": "rev",
    "natu": "nat",
    "move": "mov",
    "cana": "can",
    "euro": "eur",
    "adva": "adv",
    "medi": "med",
    "anna": "ann",
    "revi": "rev",
    "rese": "res",
    "bmj br med j": "bmj",
    "br med j": "bmj",
}


def set_container_title(records_df: pd.DataFrame) -> None:
    # Set container title
    records_df.loc[records_df[ENTRYTYPE] == ARTICLE, CONTAINER_TITLE] = records_df[
        JOURNAL
    ]
    records_df.loc[
        records_df[ENTRYTYPE].isin([INPROCEEDINGS, PROCEEDINGS, INBOOK]),
        CONTAINER_TITLE,
    ] = records_df[BOOKTITLE]
    records_df.loc[records_df[ENTRYTYPE] == BOOK, CONTAINER_TITLE] = records_df[TITLE]
    records_df[CONTAINER_TITLE] = records_df[CONTAINER_TITLE].fillna("")


def get_container_title_short(ct_array: np.array) -> np.array:
    return np.array(
        [
            "".join(item[0] for item in ct.split() if item.isalpha())
            if ct != "nan"
            else ""
            for ct in ct_array
        ]
    )


def prep_container_title(ct_array: np.array) -> np.array:
    def get_abbrev(ct: str) -> str:
        ct = str(ct)

        # Replace trailing "the" (ignore case)
        ct = re.sub(r"\sthe\s*$", "", ct, flags=re.IGNORECASE)

        ct = (
            ct.lower()
            .replace(" neuro ", " neuro")
            .replace("-", "")
            .replace("journal", "j")
        )

        if "plos one" not in ct:
            ct = parse(ct)  # replace numbers

        ct = " ".join(word[:4] for word in ct.split() if word not in JOURNAL_STOPWORDS)
        # use the same abbreviations (container_title is used for blocking)
        for original, replacement in JOURNAL_ABBREV.items():
            ct = ct.replace(original, replacement)

        # Replace trailing "supp" or leading "proc"
        ct = re.sub(r"^proc\s|\ssupp$", "", ct)
        return ct

    ct_array = np.array(
        [
            re.sub("proceedings of the", "", value.split(".")[0], flags=re.IGNORECASE)
            if "date of publication" in value.lower()
            or "conference start" in value.lower()
            else re.sub("proceedings of the", "", value, flags=re.IGNORECASE)
            for value in ct_array
        ]
    )

    def replace_journal_names(value: str) -> str:
        if (
            not any(char in value for char in "=.[")
            and len(value) < 70
            and "journal" in value.lower()
        ):
            return value
        else:
            for key, val in JOURNAL_TRANSLATIONS_DICT.items():
                if " " not in key:
                    continue

                if key in value.replace(".", " ").lower():
                    return val
            return value

    ct_array = np.array([replace_journal_names(value) for value in ct_array])

    ct_array = np.array(
        [
            re.split(r"(\.\d+)|(: )|( - )", value)[0]
            if re.search(r"(\.\d+)|(\. )|(: )|( - )", value)
            else value
            for value in ct_array
        ]
    )

    ct_array = np.array(
        [re.sub(r"\s*\[Electronic Resource\]$", "", value) for value in ct_array]
    )

    # If spaces were accidentally removed, restore them (looking for capital letters)
    # ct_array = np.array(
    #     [
    #         re.sub(r"(?<=\w)([A-Z])", r" \1", value)
    #         if "plos" not in value.lower()
    #         else value
    #         for value in ct_array
    #     ]
    # )
    # Note : distinguish conferences before (see digital_work)

    # Replace words in parentheses at the end
    ct_array = np.array(
        [re.sub(r"\s*\([^)]*\)\s*$|('s)", "", value) for value in ct_array]
    )

    # ct_array = np.array([value.replace(".", " ") for value in ct_array])
    ct_array = np.array(
        [re.sub(r"^the\s|^(l')|", "", value, flags=re.IGNORECASE) for value in ct_array]
    )
    ct_array = np.array([re.sub(r"[^A-Za-z ]+", " ", value) for value in ct_array])
    ct_array = np.array(
        [
            re.sub(r"^\s*(st|nd|rd|th) ", "", value, flags=re.IGNORECASE)
            for value in ct_array
        ]
    )

    return np.array([get_abbrev(ct) for ct in ct_array])



#! /usr/bin/env python
"""Preparation of volume field"""
import re

import numpy as np


def prep_volume(volume_array: np.array) -> np.array:
    volume_array = np.array(
        [
            re.search(r"(\d+) \(.*\)", volume).group(1)  # type: ignore
            if re.search(r"(\d+) \(.*\)", volume) is not None
            # pages included: "6 51-6"
            else re.search(r"(\d+) .*", volume).group(1)  # type: ignore
            if re.search(r"(\d+) \d+-\d+", volume) is not None
            else volume
            for volume in volume_array
        ]
    )
    volume_array = np.array(
        [
            re.search(r"(\d+) suppl \d+", volume.lower()).group(1)  # type: ignore
            if re.search(r"(\d+) suppl \d+", volume.lower()) is not None
            else volume
            for volume in volume_array
        ]
    )
    volume_array = np.array(
        [re.sub(r"[^\d\(\)]", "", volume) for volume in volume_array]
    )

    volume_array = np.array(
        [
            re.search(r"(\d+)", volume).group(0)  # type: ignore
            if re.search(r"(\d+)", volume) is not None
            else volume.replace("(", "").replace(")", "")
            for volume in volume_array
        ]
    )
    return np.array(
        [
            "" if volume == "nan" or len(volume) > 100 else volume
            for volume in volume_array
        ]
    )



#! /usr/bin/env python
"""Block for dedupe."""
import multiprocessing
import time
from datetime import datetime
from itertools import combinations

import pandas as pd

from bib_dedupe import verbose_print
from bib_dedupe.constants.fields import ABSTRACT
from bib_dedupe.constants.fields import AUTHOR_FIRST
from bib_dedupe.constants.fields import CONTAINER_TITLE_SHORT
from bib_dedupe.constants.fields import DOI
from bib_dedupe.constants.fields import NUMBER
from bib_dedupe.constants.fields import PAGES
from bib_dedupe.constants.fields import SEARCH_SET
from bib_dedupe.constants.fields import TITLE_SHORT
from bib_dedupe.constants.fields import VOLUME
from bib_dedupe.constants.fields import YEAR

block_fields_list = [
    {AUTHOR_FIRST, YEAR},
    {AUTHOR_FIRST, CONTAINER_TITLE_SHORT},
    {TITLE_SHORT, PAGES},
    {TITLE_SHORT, AUTHOR_FIRST},
    {TITLE_SHORT, VOLUME},
    {TITLE_SHORT, CONTAINER_TITLE_SHORT},
    {TITLE_SHORT, YEAR},
    {
        CONTAINER_TITLE_SHORT,
        VOLUME,
        NUMBER,
    },
    {
        CONTAINER_TITLE_SHORT,
        VOLUME,
        YEAR,
    },
    {
        CONTAINER_TITLE_SHORT,
        VOLUME,
        PAGES,
    },
    {
        CONTAINER_TITLE_SHORT,
        YEAR,
        PAGES,
    },
    {YEAR, VOLUME, NUMBER},
    {YEAR, VOLUME, PAGES},
    {YEAR, NUMBER, PAGES},
    {DOI},
    {ABSTRACT},
]


def create_pairs_for_block_fields(
    records_df: pd.DataFrame, block_fields: set
) -> pd.DataFrame:
    """
    Create pairs for block fields.

    This function creates pairs of records for the given block fields.
    It only considers records where all block fields are non-empty.

    Parameters:
    records_df (pd.DataFrame): The dataframe containing the records.
    block_fields (list): The list of block fields.

    Returns:
    pd.DataFrame: The dataframe containing the blocked pairs.
    """

    non_empty_df = records_df.loc[
        records_df[list(block_fields)].ne("").all(axis=1)
    ].copy()
    non_empty_df = non_empty_df.assign(
        block_hash=non_empty_df[list(block_fields)].apply(
            lambda x: hash(tuple(x)), axis=1
        )
    )
    if non_empty_df.empty:
        return pd.DataFrame()

    pairs = (
        non_empty_df.groupby("block_hash", group_keys=True)["ID"]
        .apply(lambda x: pd.DataFrame(list(combinations(x, 2)), columns=["ID1", "ID2"]))
        .reset_index(drop=True)
    )
    pairs["block_rule"] = "-".join(block_fields)

    pairs["require_title_overlap"] = False
    if not block_fields.intersection({TITLE_SHORT, DOI, PAGES}):
        pairs["require_title_overlap"] = True

    verbose_print.print(
        f"Blocked {str(f'{pairs.shape[0]:,}').rjust(8)} pairs with {block_fields}"
    )

    return pairs


def calculate_pairs(records_df: pd.DataFrame, block_fields: set) -> pd.DataFrame:
    """
    Calculate pairs for deduplication.

    This function calculates pairs of records for deduplication based on the given block fields.

    Parameters:
    records_df (pd.DataFrame): The dataframe containing the records.
    block_fields (list): The list of block fields.

    Returns:
    pd.DataFrame: The dataframe containing the calculated pairs.
    """

    if not all(x in records_df.columns for x in block_fields):
        return pd.DataFrame()
    pairs = create_pairs_for_block_fields(records_df, block_fields)
    return pairs


def reduce_distinct_sets(pairs_df: pd.DataFrame) -> pd.DataFrame:
    """
    Reduce distinct sets.

    This function reduces the pairs dataframe. It removes pairs where
    both records are from the same search set.

    Parameters:
    pairs_df (pd.DataFrame): The dataframe containing the pairs.

    Returns:
    pd.DataFrame: The dataframe after reduction.
    """
    if f"{SEARCH_SET}_1" not in pairs_df.columns:
        return pairs_df

    pairs_df = pairs_df[
        ~(
            (pairs_df[f"{SEARCH_SET}_1"] == pairs_df[f"{SEARCH_SET}_2"])
            & (pairs_df[f"{SEARCH_SET}_1"] != "")
        )
    ]
    return pairs_df


def reduce_non_overlapping_titles(pairs_df: pd.DataFrame) -> pd.DataFrame:
    """
    Reduce non-overlapping titles.

    This function reduces the pairs dataframe by removing pairs where the titles do not overlap.

    Parameters:
    pairs_df (pd.DataFrame): The dataframe containing the pairs.

    Returns:
    pd.DataFrame: The dataframe after reduction.
    """
    # Don't require title_overlap (words) for identical titles
    # and titles not containing a space
    pairs_df.loc[
        (pairs_df["title_1"] == pairs_df["title_2"])
        | (~pairs_df["title_1"].str.contains(" "))
        | (~pairs_df["title_2"].str.contains(" ")),
        "require_title_overlap",
    ] = False

    pairs_df = pairs_df[
        ~(
            (pairs_df["require_title_overlap"])
            & (
                pairs_df.apply(
                    lambda x: len(
                        set(str(x["title_1"]).split()) & set(str(x["title_2"]).split())
                    )
                    / (
                        # Take min because some titles have their translation appended
                        min(
                            len(str(x["title_1"]).split())
                            + 1,  # avoid division-by-zero
                            len(str(x["title_2"]).split())
                            + 1,  # avoid division-by-zero
                        )
                    )
                    < 0.5
                    if (len(str(x["title_1"])) + len(str(x["title_2"]))) != 0
                    else True,
                    axis=1,
                )
            )
        )
    ]
    return pairs_df


def block(records_df: pd.DataFrame, cpu: int = -1) -> pd.DataFrame:
    """
    Perform blocking operation on the given dataframe.

    This function performs blocking operation on the given dataframe to prepare it for deduplication.
    It calculates pairs of records for deduplication based on a predefined list of block fields.
    It then reduces the pairs dataframe by removing pairs where the titles
    do not overlap or both records are from the same search set.
    Parameters:
    records_df (pd.DataFrame): The dataframe containing the records.

    Returns:
    pd.DataFrame: The dataframe after blocking operation.
    """
    INSTRUCTION = "(please run prep(records_df) and pass the prepared df)"
    assert (
        "author_full" in records_df.columns
    ), f"Column 'author_full' not found in records_df {INSTRUCTION}"
    assert (
        "container_title" in records_df.columns
    ), f"Column 'container_title' not found in records_df {INSTRUCTION}"

    verbose_print.print(
        "Block started at " + datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    )
    start_time = time.time()

    pairs_df = pd.DataFrame(columns=["ID1", "ID2", "require_title_overlap"])
    pairs_df = pairs_df.astype({"ID1": str, "ID2": str, "require_title_overlap": bool})
    if cpu == 1:
        for field in block_fields_list:
            pairs_df = pd.concat(
                [pairs_df, calculate_pairs(records_df, field)],
                ignore_index=True,
            )

    else:
        pool = multiprocessing.Pool()
        results = pool.starmap(
            calculate_pairs, [(records_df, field) for field in block_fields_list]
        )
        pool.close()
        pool.join()

        pairs_df = pd.concat(results, ignore_index=True)

    # title overlap is only required when there is no blocked pair that requires it.
    pairs_df["require_title_overlap"] = pairs_df.groupby(["ID1", "ID2"])[
        "require_title_overlap"
    ].transform("all")
    pairs_df = pairs_df.drop_duplicates(subset=["ID1", "ID2"])

    pairs_df = pd.merge(
        pairs_df,
        records_df.add_suffix("_1"),
        left_on="ID1",
        right_on="ID_1",
        how="left",
        suffixes=("", "_1"),
    )

    pairs_df = pd.merge(
        pairs_df,
        records_df.add_suffix("_2"),
        left_on="ID2",
        right_on="ID_2",
        how="left",
        suffixes=("", "_2"),
    )

    verbose_print.print(f"Blocked {str(f'{pairs_df.shape[0]:,}').rjust(8)} pairs")

    pairs_df = reduce_non_overlapping_titles(pairs_df)
    pairs_df = reduce_distinct_sets(pairs_df)

    verbose_print.print(f"Blocked pairs reduced to {pairs_df.shape[0]:,} pairs")
    end_time = time.time()
    verbose_print.print(f"Block completed after: {end_time - start_time:.2f} seconds")
    return pairs_df



#! /usr/bin/env python
"""Match for dedupe"""
import time
from datetime import datetime

import pandas as pd

import bib_dedupe.match_conditions
import bib_dedupe.sim
import bib_dedupe.util
from bib_dedupe import verbose_print
from bib_dedupe.constants.colors import END
from bib_dedupe.constants.colors import GREEN
from bib_dedupe.constants.colors import ORANGE
from bib_dedupe.constants.colors import RED
from bib_dedupe.constants.fields import ABSTRACT
from bib_dedupe.constants.fields import AUTHOR
from bib_dedupe.constants.fields import CONTAINER_TITLE
from bib_dedupe.constants.fields import DOI
from bib_dedupe.constants.fields import DUPLICATE
from bib_dedupe.constants.fields import DUPLICATE_LABEL
from bib_dedupe.constants.fields import ID
from bib_dedupe.constants.fields import MAYBE
from bib_dedupe.constants.fields import NUMBER
from bib_dedupe.constants.fields import PAGE_RANGES_ADJACENT
from bib_dedupe.constants.fields import PAGES
from bib_dedupe.constants.fields import SEARCH_SET
from bib_dedupe.constants.fields import TITLE
from bib_dedupe.constants.fields import VOLUME
from bib_dedupe.constants.fields import YEAR

SIM_FIELDS_FLOAT = [
    AUTHOR,
    TITLE,
    CONTAINER_TITLE,
    YEAR,
    VOLUME,
    NUMBER,
    PAGES,
    ABSTRACT,
    DOI,
]

SIM_FIELDS = SIM_FIELDS_FLOAT + [PAGE_RANGES_ADJACENT]

NON_DUPLICATE_CONDITIONS = bib_dedupe.match_conditions.non_duplicate_conditions
DUPLICATE_CONDITIONS = bib_dedupe.match_conditions.duplicate_conditions


def __print_details(pairs: pd.DataFrame) -> None:
    if verbose_print.verbosity_level < 2:
        return

    verbose_print.pretty_print(pairs.iloc[0].to_dict())
    if pairs.shape[0] == 0:
        return

    for item in SIM_FIELDS:
        similarity = pairs.loc[0, item]
        if item == PAGE_RANGES_ADJACENT:
            verbose_print.print(f"{RED}{PAGE_RANGES_ADJACENT}: {similarity}{END}")
            continue
        if similarity > 0.9:
            verbose_print.print(f"{GREEN}Similarity {item:<20}: {similarity:.2f}{END}")
        elif similarity > 0.6:
            verbose_print.print(f"{ORANGE}Similarity {item:<20}: {similarity:.2f}{END}")
        else:
            verbose_print.print(f"{RED}Similarity {item:<20}: {similarity:.2f}{END}")

    verbose_print.print("Merge conditions that matched:")
    for duplicate_condition in DUPLICATE_CONDITIONS:
        if pairs.query(duplicate_condition, engine="python").shape[0] > 0:
            verbose_print.print(f"{GREEN}{duplicate_condition}{END}")
        else:
            verbose_print.print(f"{duplicate_condition}")

    verbose_print.print()
    verbose_print.print(pairs)
    verbose_print.print("Non-merge conditions that matched:")
    for non_duplicate_condition in NON_DUPLICATE_CONDITIONS:
        if pairs.query(non_duplicate_condition, engine="python").shape[0] != 0:
            verbose_print.print(f"{RED}{non_duplicate_condition}{END}")
        else:
            verbose_print.print(non_duplicate_condition)


def __get_true_pairs(pairs: pd.DataFrame) -> pd.DataFrame:
    true_pairs = pairs.query("|".join(DUPLICATE_CONDITIONS), engine="python")
    true_pairs = true_pairs.query(
        "~(" + " | ".join(NON_DUPLICATE_CONDITIONS) + ")", engine="python"
    )
    true_pairs = true_pairs.drop_duplicates()

    # Add a label column to each dataframe
    true_pairs[DUPLICATE_LABEL] = DUPLICATE
    # Select the ID_1, SEARCH_SET_1 and ID_2, SEARCH_SET_2 fields and the new label column
    true_pairs = true_pairs[
        [f"{ID}_1", f"{SEARCH_SET}_1", f"{SEARCH_SET}_2", f"{ID}_2", DUPLICATE_LABEL]
    ]

    return true_pairs


def __get_maybe_pairs(pairs: pd.DataFrame, true_pairs: pd.DataFrame) -> pd.DataFrame:
    maybe_pairs = pd.DataFrame()

    maybe_pairs = pairs[
        (pairs[TITLE] > 0.85) & (pairs["author"] > 0.75)
        | (pairs[TITLE] > 0.8) & (pairs[ABSTRACT] > 0.8)
        | (pairs[TITLE] > 0.8) & (pairs[CONTAINER_TITLE] > 0.8)
        | (pd.isna(pairs[DOI]) | (pairs[DOI] > 0.99) | (pairs[DOI] == 0))
        & ~(
            (
                pd.to_numeric(pairs[f"{YEAR}_1"], errors="coerce")
                - pd.to_numeric(pairs[f"{YEAR}_2"], errors="coerce")
                > 1
            )
            | (
                pd.to_numeric(pairs[f"{YEAR}_2"], errors="coerce")
                - pd.to_numeric(pairs[f"{YEAR}_1"], errors="coerce")
                > 1
            )
        )
    ]

    maybe_pairs = maybe_pairs.drop_duplicates()

    # Drop from maybe_pairs where ID_1 ID_2 combinations are in true_pairs
    maybe_pairs = maybe_pairs[
        ~maybe_pairs.set_index([f"{ID}_1", f"{ID}_2"]).index.isin(
            true_pairs.set_index([f"{ID}_1", f"{ID}_2"]).index
        )
    ]

    # Add a label column to each dataframe
    maybe_pairs[DUPLICATE_LABEL] = MAYBE
    # Select the ID_1, SEARCH_SET_1 and ID_2, SEARCH_SET_2 fields and the new label column
    maybe_pairs = maybe_pairs[
        [f"{ID}_1", f"{SEARCH_SET}_1", f"{SEARCH_SET}_2", f"{ID}_2", DUPLICATE_LABEL]
    ]

    return maybe_pairs


def match(pairs: pd.DataFrame, cpu: int = -1) -> pd.DataFrame:
    pairs = bib_dedupe.sim.calculate_similarities(pairs, cpu=cpu)

    verbose_print.print(
        "Match started at " + datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    )
    start_time = time.time()

    if pairs.empty:
        end_time = time.time()
        verbose_print.print(
            f"Match completed after: {end_time - start_time:.2f} seconds"
        )

        return pd.DataFrame(columns=[f"{ID}_1", f"{ID}_2", DUPLICATE_LABEL])

    for field in SIM_FIELDS_FLOAT:
        pairs[field] = pairs[field].astype(float)

    remaining_fields = set(pairs.columns) - set(SIM_FIELDS_FLOAT)
    for field in remaining_fields:
        pairs[field] = pairs[field].astype(str)

    __print_details(pairs)

    true_pairs = __get_true_pairs(pairs)

    maybe_pairs = __get_maybe_pairs(pairs, true_pairs)

    end_time = time.time()
    verbose_print.print(f"Match completed after: {end_time - start_time:.2f} seconds")

    return pd.concat([true_pairs, maybe_pairs])



#! /usr/bin/env python
"""Similarities for dedupe"""
import concurrent.futures
import re
import time
from datetime import datetime

import numpy as np
import pandas as pd
from rapidfuzz import fuzz

from bib_dedupe import verbose_print
from bib_dedupe.constants.fields import ABSTRACT
from bib_dedupe.constants.fields import AUTHOR
from bib_dedupe.constants.fields import CONTAINER_TITLE
from bib_dedupe.constants.fields import DOI
from bib_dedupe.constants.fields import ID
from bib_dedupe.constants.fields import NUMBER
from bib_dedupe.constants.fields import PAGE_RANGES_ADJACENT
from bib_dedupe.constants.fields import PAGES
from bib_dedupe.constants.fields import TITLE
from bib_dedupe.constants.fields import VOLUME
from bib_dedupe.constants.fields import YEAR

TITLE_STOPWORDS = ["the", "a", "an", "in", "on", "at", "and", "or", "of"]


def sim_token_sort_ratio(string_1: str, string_2: str) -> float:
    """Calculate the token sort ratio similarity between two strings."""
    if string_1 != "" and string_2 != "":
        return fuzz.token_sort_ratio(string_1, string_2) / 100
    return 0.0


def sim_author(
    author_1: str,
    author_full_1: str,
    author_2: str,
    author_full_2: str,
) -> float:
    """
    Calculate the author similarity between two strings.
    """

    if author_1 == "" and author_2 == "":
        return 0.0

    if author_1 == author_2:  # to save computational time
        return 1.0

    author_1 = author_1.replace("vander", "")
    author_2 = author_2.replace("vander", "")

    abbreviated = (
        fuzz.token_sort_ratio(author_1[:200], author_2[:200]) / 100
        if len(author_1) > 200 or len(author_2) > 200
        else 0
    )
    author_partial_diff = fuzz.partial_ratio(author_1, author_2) / 100

    author_diff = 0.0

    if len(author_full_1) > 5 and len(author_full_2) > 5:
        # Extract capital letters from both author_full_1 and author_full_2
        capital_letters_author_full_1 = re.findall(r"[A-Z]", author_full_1)
        capital_letters_author_full_2 = re.findall(r"[A-Z]", author_full_2)
        # Calculate similarity of capital letters in author_full_1 and author_full_2
        capital_letters = (
            fuzz.token_sort_ratio(
                " ".join(capital_letters_author_full_1),
                " ".join(capital_letters_author_full_2),
            )
            / 100
        )
        author_diff = capital_letters

    author_full_diff = 0.0

    if author_full_1 != "" and author_full_2 != "":
        author_full_diff = fuzz.token_sort_ratio(author_full_1, author_full_2) / 100

    return max(abbreviated, author_partial_diff, author_diff, author_full_diff)


def sim_page(pages_string_1: str, pages_string_2: str) -> float:
    """
    Calculate the page similarity between two strings.
    """

    pages_string_1 = re.sub(r"[a-zA-Z]", "", pages_string_1)
    pages_string_2 = re.sub(r"[a-zA-Z]", "", pages_string_2)

    if pages_string_1 == "" and pages_string_2 == "":
        return 0.0

    if (
        (
            (pages_string_1.endswith(pages_string_2) and pages_string_2.isdigit())
            or (pages_string_2.endswith(pages_string_1) and pages_string_1.isdigit())
        )
        and pages_string_1 != ""
        and pages_string_2 != ""
    ):
        return 1.0

    pages_string_1_match = re.search(r"\d+", pages_string_1)
    pages_string_1 = pages_string_1_match.group() if pages_string_1_match else ""
    pages_string_2_match = re.search(r"\d+", pages_string_2)
    pages_string_2 = pages_string_2_match.group() if pages_string_2_match else ""

    if (
        pages_string_1 != ""
        and pages_string_2 != ""
        and pages_string_1 == pages_string_2
    ):
        return 1
    else:
        return fuzz.token_sort_ratio(pages_string_1, pages_string_2) / 100


def sim_title(title_1: str, title_2: str, debug: bool = False) -> float:
    """
    Calculate the title similarity between two strings.
    """

    t1 = str(title_1)
    t2 = str(title_2)

    if t1 == "" and t2 == "":
        return 0.0

    if t1.replace(" ", "") == t2.replace(" ", "") and t1.replace(" ", "") != "":
        return 1.0

    # comment/reply etc.
    if any(
        [
            ((term in t1 and term not in t2) or (term in t2 and term not in t1))
            for term in [
                "comment",
                "response",
                "reply",
            ]
        ]
    ):
        return 0

    # In long titles, secondary titles may not be added in all cases
    if len(t1) > 60 and len(t2) > 60 and t1.startswith(t2) or t2.startswith(t1):
        return 1.0

    # Remove chemical formulae
    if "[" in t1:
        t1 = re.sub(r"\[[a-z0-9 ]{1,5}\]", "", t1)
    if "[" in t2:
        t2 = re.sub(r"\[[a-z0-9 ]{1,5}\]", "", t2)

    # Similarity for mismatching numbers (part 1 vs 2) should be 0
    t1_digits = re.findall(r"(?<!\[)\d+", t1)
    t2_digits = re.findall(r"(?<!\[)\d+", t2)
    if (
        t1_digits != t2_digits
        and "".join(t1_digits) != "".join(t2_digits)
        and len(t1_digits) < 3
    ):
        if debug:
            verbose_print.print(f"mismatching digits: {t1_digits} - {t2_digits}")
        return 0

    # Women vs men split studies
    if any(
        [
            ((term in t1 and term not in t2) or (term in t2 and term not in t1))
            for term in [
                "women",
                "adult",
                "pediatric",
                "protocol",
                "vivo",
                "vitro",
                "rats",
                "cats",
            ]
        ]
    ):
        return 0

    if any(term in t1 for term in ["part", "effect", "treatment"]):
        t1_parts = re.findall(r"part [a-z]", t1)
        t2_parts = re.findall(r"part [a-z]", t2)
        if t1_parts != t2_parts and len(t1_digits) < 3:
            return 0

        effect_1 = re.findall(r"effect[s]? of (\w+)", t1)
        effect_2 = re.findall(r"effect[s]? of (\w+)", t2)
        if effect_1 != effect_2:
            return 0

        treatment_1 = re.findall(r"treatment of (\w+)", t1)
        treatment_2 = re.findall(r"treatment of (\w+)", t2)
        if treatment_1 != treatment_2:
            return 0

        patients_1 = re.findall(r"(\w+) patients", t1)
        patients_2 = re.findall(r"(\w+) patients", t2)
        if patients_1 != patients_2:
            return 0

    if t1.endswith(t2) or t2.endswith(t1):
        return 1.0

    # Remove common stopwords from t1 and t2
    t1 = " ".join(word for word in t1.split() if word not in TITLE_STOPWORDS)
    t2 = " ".join(word for word in t2.split() if word not in TITLE_STOPWORDS)

    # Insert a space between digits that come directly after a string
    t1 = re.sub(r"([A-Za-z])(\d)", r"\1 \2", t1)
    t2 = re.sub(r"([A-Za-z])(\d)", r"\1 \2", t2)

    title_diff = fuzz.ratio(t1, t2) / 100

    # Title fields containing translated version of the title
    if title_diff < 0.7:
        if len(t1) > (1.7 * len(t2)):
            partial_ratio = fuzz.partial_ratio(t1, t2) / 100
            if partial_ratio > title_diff:
                return partial_ratio
        if len(t2) > (1.7 * len(t1)):
            partial_ratio = fuzz.partial_ratio(t2, t1) / 100
            if partial_ratio > title_diff:
                return partial_ratio

    return title_diff


def sim_year(year_1_str: str, year_2_str: str) -> float:
    """
    Calculate the similarity between two years.
    """

    # Convert years to integers, invalid conversions will be replaced with -1
    def convert_to_int(year_str: str) -> int:
        if year_str.isdigit():
            return int(year_str)
        else:
            return -1

    year_1 = convert_to_int(year_1_str)
    year_2 = convert_to_int(year_2_str)

    # Calculate absolute difference between the years
    year_diff = abs(year_1 - year_2)

    # Calculate similarity based on the difference
    similarity = (
        1.0
        if year_diff == 0
        else 0.95
        if year_diff == 1
        else 0.8
        if year_diff == 2
        else 0
    )

    # Set similarity to 0 where any of the years is -1
    similarity = 0 if year_1 == -1 or year_2 == -1 else similarity

    return similarity


def sim_doi(doi_1_str: str, doi_2_str: str) -> float:
    if doi_1_str == "" or doi_2_str == "":
        return 0

    return fuzz.ratio(doi_1_str, doi_2_str) / 100


def sim_number(n1_str: str, n2_str: str) -> float:
    """
    Calculate the similarity between two numbers.
    """

    if n1_str == "" or n2_str == "":
        return 0

    n1 = int(n1_str) if str(n1_str).isdigit() else 0
    n2 = int(n2_str) if str(n2_str).isdigit() else 0
    if n1 > 12 and n2 > 12:
        number_diff = abs(n1 - n2)
        if number_diff == 0:
            return 1.0
        elif number_diff <= 2:
            return 0.95
        else:
            return 0.0
    else:
        # some journals have numbers like 3/4 (which can be abbreviated)
        return fuzz.token_sort_ratio(str(n1), str(n2)) / 100


def sim_volume(v1_str: str, v2_str: str) -> float:
    """
    Calculate the similarity between two volumes.
    """

    if v1_str == "" or v2_str == "":
        return 0

    v1 = int(v1_str) if str(v1_str).isdigit() else 0
    v2 = int(v2_str) if str(v2_str).isdigit() else 0

    if v1 == v2:
        return 1.0
    return 0.0


def sim_abstract(abstract_1: str, abstract_2: str) -> float:
    abstract_1 = str(abstract_1)
    abstract_2 = str(abstract_2)

    if abstract_1 == "" or abstract_2 == "":
        return 0.0

    if len(abstract_1) > 500 and len(abstract_2) > 500:
        if abstract_1.startswith(abstract_2[:-100]) or abstract_2.startswith(
            abstract_1[:-100]
        ):
            return 1.0

    return fuzz.ratio(abstract_1, str(abstract_2)) / 100


def sim_container_title(container_1: str, container_2: str) -> float:
    """
    Calculate the similarity between two containers.
    """

    if container_1 == "" or container_2 == "":
        return 0.0

    if ("euro " in container_1 and "am " in container_2) or (
        "euro " in container_2 and "am " in container_1
    ):
        return 0.0

    container_1 = container_1.replace("res", "")
    container_2 = container_2.replace("res", "")

    abbreviation_match = 0
    if " " not in container_1 and " " in container_2:
        first_letters_1 = container_1
        first_letters_2 = "".join(word[0] for word in container_2.split())
        abbreviation_match = 1 if first_letters_1 == first_letters_2 else 0
    if " " not in container_2 and " " in container_1:
        first_letters_1 = "".join(word[0] for word in container_1.split())
        first_letters_2 = container_2
        abbreviation_match = 1 if first_letters_1 == first_letters_2 else 0
    word_match = 1
    words_1 = container_1.split()
    words_2 = container_2.split()
    if len(words_1) != len(words_2):
        word_match = 0
    else:
        for word_1, word_2 in zip(words_1, words_2):
            if not word_1.startswith(word_2) and not word_2.startswith(word_1):
                word_match = 0
                break

    spaces_1 = container_1.count(" ")
    spaces_2 = container_2.count(" ")

    if spaces_1 < 5 and spaces_2 < 5:
        if container_1.startswith(container_2) or container_2.startswith(container_1):
            return 1.0
        overall = fuzz.ratio(str(container_1), str(container_2)) / 100
    else:
        overall = fuzz.partial_ratio(str(container_1), str(container_2)) / 100

    return max(overall, abbreviation_match, word_match)


def page_ranges_adjacent(row: pd.Series) -> str:
    """
    This function checks if the page ranges of two articles are adjacent.

    Parameters:
    row (pd.Series): A row of a DataFrame containing the page ranges of two articles.

    Returns:
    float: Returns 1.0 if the page ranges are adjacent, 0.0 otherwise.
    """
    if row["pages_1"] == row["pages_2"]:
        return ""
    if not re.match(r"\d{1,}-\d{1,}", row["pages_1"]) or not re.match(
        r"\d{1,}-\d{1,}", row["pages_2"]
    ):
        return ""

    pages_1 = row["pages_1"].split("-")
    pages_2 = row["pages_2"].split("-")
    if len(pages_1) == 2 and len(pages_2) == 2:
        pages_1_start, pages_1_end = map(int, pages_1)

        pages_2_start, pages_2_end = map(int, pages_2)

        if pages_1_end + 1 == pages_2_start or pages_2_end + 1 == pages_1_start:
            return "adjacent"

        if pages_1_end < pages_2_start or pages_2_end < pages_1_start:
            return "non_overlapping"

    return ""


similarity_functions = {
    PAGES: sim_page,
    TITLE: sim_title,
    YEAR: sim_year,
    NUMBER: sim_number,
    CONTAINER_TITLE: sim_container_title,
    VOLUME: sim_volume,
    ABSTRACT: sim_abstract,
    DOI: sim_doi,
}


def process_df_split(split_df: pd.DataFrame) -> pd.DataFrame:
    for index, row in split_df.iterrows():
        split_df.loc[index, AUTHOR] = sim_author(
            str(row["author_1"]),
            str(row["author_full_1"]),
            str(row["author_2"]),
            str(row["author_full_2"]),
        )

        for field, function in similarity_functions.items():
            split_df.loc[index, field] = function(  # type: ignore
                str(row[f"{field}_1"]), str(row[f"{field}_2"])
            )

        split_df.loc[index, PAGE_RANGES_ADJACENT] = page_ranges_adjacent(row)

    return split_df


def calculate_similarities(pairs_df: pd.DataFrame, cpu: int) -> pd.DataFrame:
    """
    Calculate similarities between pairs of data in a DataFrame.

    Args:
        pairs_df (pd.DataFrame): DataFrame containing pairs of data.

    Returns:
        pd.DataFrame: DataFrame with calculated similarities.
    """

    required_columns = [
        f"{col}_1"
        for col in [
            ID,
            "author_full",
            TITLE,
            YEAR,
            ABSTRACT,
            AUTHOR,
            CONTAINER_TITLE,
            DOI,
            NUMBER,
            PAGES,
            VOLUME,
        ]
    ] + [
        f"{col}_2"
        for col in [
            ID,
            "author_full",
            TITLE,
            YEAR,
            ABSTRACT,
            AUTHOR,
            CONTAINER_TITLE,
            DOI,
            NUMBER,
            PAGES,
            VOLUME,
        ]
    ]
    missing_columns = list(
        filter(lambda col: col not in pairs_df.columns, required_columns)
    )
    if missing_columns:
        raise ValueError(f"Missing required columns: {', '.join(missing_columns)}")

    verbose_print.print(
        "Sim started at " + datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    )
    start_time = time.time()
    if cpu == 1:
        pairs_df = process_df_split(pairs_df)
    else:
        df_split = np.array_split(pairs_df, 8)

        with concurrent.futures.ProcessPoolExecutor(max_workers=8) as executor:
            results = executor.map(process_df_split, df_split)

        pairs_df = pd.concat(list(results))

    end_time = time.time()
    verbose_print.print(f"Sim completed after: {end_time - start_time:.2f} seconds")

    return pairs_df



#! /usr/bin/env python
"""Preparation of author field"""
import re
import typing
import unicodedata
from typing import List

import numpy as np

# Note: for re.sub, "van der" must be checked before "van"
NAME_PREFIXES_LOWER = [
    "van der",
    "van",
    "von der",
    "von",
    "vom",
    "le",
    "den",
    "der",
    "ter",
    "de",
    "da",
    "di",
]


def get_author_format_case(authors: typing.List[str], original_string: str) -> str:
    if authors == [""]:
        return "empty"

    if any(
        keyword in original_string.lower() for keyword in ["group", "agency", "council"]
    ):
        return "organization"

    if (" and " in original_string and ", " in original_string) or (
        " and " not in original_string
        and ", " in original_string
        and len(original_string) < 50
    ):
        return "proper_format"

    if len(authors) < 4 and not any("," in a for a in authors):
        return "single_author_missing_comma"

    if " and " not in authors and "," not in authors and len(original_string) > 5:
        if (
            sum(
                1
                for author in authors
                if author.isupper() and len(re.sub(r"[ .-]", "", author)) < 3
            )
            / len(authors)
            >= 0.4
        ):
            return "abbreviated_initials"

        if (
            sum(1 for match in re.findall(r"[A-Z][a-z\.]+[A-Z][a-z]+", original_string))
            / len(authors)
            >= 0.1
        ):
            if " " in original_string:
                return "missing_spaces_between_words"
            else:
                return "no_spaces_at_al"

    return "special_case"


def get_authors_split(authors: str) -> list:
    # single author
    if len(authors) < 15:
        if " " not in authors and re.search(r"[a-z]{3}[A-Z]", authors):
            return re.split(r"(?<=[a-z])(?=[A-Z])", authors)

        if authors.count(" ") <= 2:
            return authors.split(" ")

    authors_list = re.split(r"(?=[A-Z])", authors)
    for i in range(len(authors_list) - 1):
        if (
            authors_list[i].endswith("-")
            or authors_list[i] in ["Mc", "Mac"]
            or (
                len(authors_list[i]) == 1
                and authors_list[i].isupper()
                and len(authors_list[i + 1]) == 1
                and authors_list[i + 1].isupper()
            )
        ):
            authors_list[i + 1] = authors_list[i] + authors_list[i + 1]
            authors_list[i] = ""
    return [author.rstrip() for author in authors_list if author != ""]


def remove_accents(*, input_str: str) -> str:
    """Replace the accents in a string"""

    def rmdiacritics(char: str) -> str:
        """
        Return the base character of char, by "removing" any
        diacritics like accents or curls and strokes and the like.
        """
        try:
            desc = unicodedata.name(char)
            cutoff = desc.find(" WITH ")
            if cutoff != -1:
                desc = desc[:cutoff]
                char = unicodedata.lookup(desc)
        except (KeyError, ValueError):
            pass  # removing "WITH ..." produced an invalid name
        return char

    try:
        nfkd_form = unicodedata.normalize("NFKD", input_str)
        wo_ac_list = [
            rmdiacritics(c) for c in nfkd_form if not unicodedata.combining(c)
        ]
        wo_ac = "".join(wo_ac_list)
    except ValueError:
        wo_ac = input_str
    return wo_ac


def __prep_author_abbreviated_initials(authors_list: typing.List[str]) -> str:
    temp_authors_list: typing.List[str] = []
    next_author: typing.List[str] = []
    for author_fragment in authors_list:
        if re.match(r"[A-Z][a-z]+", author_fragment):
            temp_authors_list.append(" ".join(next_author))
            next_author = [author_fragment]
        else:
            next_author.append(author_fragment)
    temp_authors_list.append(" ".join(next_author))
    temp_authors_list = [author for author in temp_authors_list if author != ""]

    for i in range(len(temp_authors_list)):
        words = temp_authors_list[i].split()
        for j in range(len(words) - 1, -1, -1):
            if words[j].isupper() and not words[j - 1].isupper():
                words[j - 1] = words[j - 1] + ","
                break
        temp_authors_list[i] = " ".join(words)

    for i in range(len(temp_authors_list)):
        if i == len(temp_authors_list) - 1:
            temp_authors_list[i] = temp_authors_list[i]
        elif ", " in temp_authors_list[i]:
            temp_authors_list[i] = temp_authors_list[i] + " and "
        else:
            temp_authors_list[i] = temp_authors_list[i] + " "

    authors_str = "".join(temp_authors_list)
    return authors_str


def __prep_author_no_spaces_at_al(authors_list: List[str]) -> str:
    authors_str = ""
    for element in authors_list:
        if re.match(r"^[A-Z][a-z]+", element):
            authors_str += element + " "
        else:
            authors_str += ", " + element + " and "
    authors_str = authors_str.rstrip(" and ")
    return authors_str


def __prep_author_missing_spaces_between_words(original_string: str) -> str:
    new_authors = re.sub(
        r"([A-Z][a-z.]+)([A-Z])", r"\1 SPLIT\2", original_string
    ).split("SPLIT")
    for i, element in enumerate(new_authors):
        words = [
            w.replace(".", "").rstrip()
            for w in element.split()
            if w.lower() not in NAME_PREFIXES_LOWER
        ]
        if len(words) > 1:
            middle_index = len(words) // 2
            words.insert(middle_index, ",")
            new_authors[i] = " ".join(words)
    authors_str = " and ".join(new_authors)
    return authors_str


def preprocess_author(authors: str, *, debug: bool) -> str:
    authors = str(authors)
    if authors.lower() in ["nan", "anonymous"]:
        return ""

    # Many databases do not handle accents properly...
    authors = (
        remove_accents(input_str=authors)
        .replace("ue", "u")
        .replace("oe", "o")
        .replace("ae", "a")
    )

    authors = re.sub(r"\d", "", authors)

    if ";" in authors:
        authors = authors.replace(";", " and ")

    # Capitalize von/van/... and add "-" to connect last-names
    authors = re.sub(
        r"([A-Z])(" + "|".join(NAME_PREFIXES_LOWER) + r") (\w+)",
        lambda match: match.group(1)
        + match.group(2).title().replace(" ", "-")
        + "-"
        + match.group(3),
        authors,
    )
    authors = re.sub(
        r"(^| |\.|-)(" + "|".join(NAME_PREFIXES_LOWER) + r") (\w+)",
        lambda match: match.group(1)
        + match.group(2).title().replace(" ", "-")
        + "-"
        + match.group(3),
        authors,
        flags=re.IGNORECASE,
    )

    original_string = authors

    authors_list = get_authors_split(authors)

    format_case = get_author_format_case(authors_list, original_string)

    # TODO : handle debug / verbose_print
    if debug:
        print(original_string)
        print(f"format_case: {format_case}")
        print(f"authors_list: {authors_list}")

    if format_case == "proper_format":
        authors_str = authors

    elif format_case == "organization":
        authors_str = authors

    elif format_case == "empty":
        authors_str = ""

    elif format_case == "single_author_missing_comma":
        if authors_list[0].isupper():
            authors_list[0] = authors_list[0].title()

        authors_str = authors_list[0] + ", " + " ".join(authors_list[1:])

    # Broadley K.Burton A. C.Avgar T.Boutin S.
    elif format_case == "abbreviated_initials":
        authors_str = __prep_author_abbreviated_initials(authors_list)

    # PayenJ.-L.IzopetJ.Galindo-MigeotV.Lauwers-Cances
    elif format_case == "no_spaces_at_al":
        authors_str = __prep_author_no_spaces_at_al(authors_list)

    # Vernia FilippoDi Ruscio MirkoStefanelli GianpieroViscido AngeloFrieri GiuseppeLatella Giovanni
    elif format_case == "missing_spaces_between_words":
        authors_str = __prep_author_missing_spaces_between_words(original_string)

    else:
        # For debugging:
        # print(f"{format_case}: {original_string}")
        authors_str = " and ".join(authors_list)

    authors_str = authors_str.replace(" ,", ",")
    authors_str = re.sub(r"[^A-Za-z0-9, ]+", "", authors_str)
    return authors_str.lower()


def prep_authors(authors_array: np.array, *, debug: bool = False) -> np.array:
    return np.array(
        [preprocess_author(author, debug=debug) for author in authors_array]
    )


def select_authors(authors_array: np.array) -> np.array:
    def select_author(authors: str) -> str:
        """Select first author"""

        authors_list = authors.split(" and ")
        authors_str = " ".join(
            [
                re.sub(
                    r"(^| )(" + "|".join(NAME_PREFIXES_LOWER) + r") ",
                    lambda match: match.group(1) + match.group(2).replace(" ", "-") + "-",  # type: ignore
                    author.split(",")[0],
                    flags=re.IGNORECASE,
                ).replace(" ", "")
                for author in authors_list
            ][:8]
        )
        authors_str = authors_str.replace("anonymous", "").replace("jr", "")
        authors_str = re.sub(r"[^A-Za-z0-9, ]+", "", authors_str)
        return authors_str

    return np.array([select_author(author) for author in authors_array])



#! /usr/bin/env python
"""Preparation of pages field"""
import re

import numpy as np

month_dict = {
    "jan": "01",
    "feb": "02",
    "mar": "03",
    "apr": "04",
    "may": "05",
    "jun": "06",
    "jul": "07",
    "aug": "08",
    "sep": "09",
    "oct": "10",
    "nov": "11",
    "dec": "12",
}


def prep_pages(pages_array: np.array) -> np.array:
    def process_page(value: str) -> str:
        if value.isalpha():
            return ""

        # Fix Excel errors (e.g., "08-11" is converted to "08-Nov")
        for month, num in month_dict.items():
            if month in value.lower():
                value = value.lower().replace(month, num)
                break

        def roman_to_int(s: str) -> int:
            rom_val = {"I": 1, "V": 5, "X": 10, "L": 50, "C": 100, "D": 500, "M": 1000}
            int_val = 0
            for i in range(len(s)):
                if i > 0 and rom_val[s[i]] > rom_val[s[i - 1]]:
                    int_val += rom_val[s[i]] - 2 * rom_val[s[i - 1]]
                else:
                    int_val += rom_val[s[i]]
            return int_val

        # Check if the value is a roman numeral range
        roman_numeral_range = re.match(r"([IVXLCDM]+)-([IVXLCDM]+)", value, re.I)
        if roman_numeral_range:
            # Convert the roman numerals to integers
            from_page = roman_to_int(roman_numeral_range.group(1).upper())
            to_page = roman_to_int(roman_numeral_range.group(2).upper())
            return f"{from_page}-{to_page}"

        # Remove leading zeros in groups of digits
        value = re.sub(r"\b0+([0-9]+)", r"\1", value)

        value = re.sub(r"[A-Za-z. ]*", "", value)
        if re.match(r"^\d+\s*-?-\s*\d+$", value):
            from_page_str, to_page_str = re.findall(r"(\d+)", value)
            if from_page_str == to_page_str:
                return from_page_str
            if len(from_page_str) > len(to_page_str):
                return (
                    f"{from_page_str}-{from_page_str[:-len(to_page_str)]}{to_page_str}"
                )
            else:
                return f"{from_page_str}-{to_page_str}"
        if value in [
            " ",
            None,
            "nan",
            "na",
            "no pages",
            "no pagination",
            "var.pagings",
        ]:
            return ""
        # re sub everything except numbers and dashes
        return re.sub(r"[^0-9-]", "", value).lstrip("-").rstrip("-")

    return np.vectorize(process_page)(pages_array)
